# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""LogPane class."""

import functools
import logging
import re
import time
from typing import (
    Any,
    Callable,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
)

from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import (
    Condition,
    has_focus,
)
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.key_binding import (
    KeyBindings,
    KeyPressEvent,
    KeyBindingsBase,
)
from prompt_toolkit.layout import (
    ConditionalContainer,
    Float,
    FloatContainer,
    FormattedTextControl,
    HSplit,
    UIContent,
    UIControl,
    VerticalAlign,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType, MouseButton

from pw_console.log_view import LogView
from pw_console.log_pane_toolbars import (
    LineInfoBar,
    TableToolbar,
)
from pw_console.log_pane_saveas_dialog import LogPaneSaveAsDialog
from pw_console.log_pane_selection_dialog import LogPaneSelectionDialog
from pw_console.log_store import LogStore
from pw_console.search_toolbar import SearchToolbar
from pw_console.filter_toolbar import FilterToolbar

from pw_console.style import (
    get_pane_style,
)
from pw_console.widgets import (
    ToolbarButton,
    WindowPane,
    WindowPaneHSplit,
    WindowPaneToolbar,
    create_border,
    mouse_handlers,
    to_checkbox_text,
    to_keybind_indicator,
)


if TYPE_CHECKING:
    from pw_console.console_app import ConsoleApp

_LOG_OUTPUT_SCROLL_AMOUNT = 5
_LOG = logging.getLogger(__package__)


class LogContentControl(UIControl):
    """LogPane prompt_toolkit UIControl for displaying LogContainer lines."""

    def __init__(self, log_pane: 'LogPane') -> None:
        # pylint: disable=too-many-locals
        self.log_pane = log_pane
        self.log_view = log_pane.log_view

        # Mouse drag visual selection flags.
        self.visual_select_mode_drag_start = False
        self.visual_select_mode_drag_stop = False

        self.uicontent: Optional[UIContent] = None
        self.lines: List[StyleAndTextTuples] = []

        # Key bindings.
        key_bindings = KeyBindings()
        register = log_pane.application.prefs.register_keybinding

        @register('log-pane.shift-line-to-top', key_bindings)
        def _shift_log_to_top(_event: KeyPressEvent) -> None:
            """Shift the selected log line to the top."""
            self.log_view.move_selected_line_to_top()

        @register('log-pane.shift-line-to-center', key_bindings)
        def _shift_log_to_center(_event: KeyPressEvent) -> None:
            """Shift the selected log line to the center."""
            self.log_view.center_log_line()

        @register('log-pane.toggle-wrap-lines', key_bindings)
        def _toggle_wrap_lines(_event: KeyPressEvent) -> None:
            """Toggle log line wrapping."""
            self.log_pane.toggle_wrap_lines()

        @register('log-pane.toggle-table-view', key_bindings)
        def _toggle_table_view(_event: KeyPressEvent) -> None:
            """Toggle table view."""
            self.log_pane.toggle_table_view()

        @register('log-pane.duplicate-log-pane', key_bindings)
        def _duplicate(_event: KeyPressEvent) -> None:
            """Duplicate this log pane."""
            self.log_pane.duplicate()

        @register('log-pane.remove-duplicated-log-pane', key_bindings)
        def _delete(_event: KeyPressEvent) -> None:
            """Remove log pane."""
            if self.log_pane.is_a_duplicate:
                self.log_pane.application.window_manager.remove_pane(
                    self.log_pane
                )

        @register('log-pane.clear-history', key_bindings)
        def _clear_history(_event: KeyPressEvent) -> None:
            """Clear log pane history."""
            self.log_pane.clear_history()

        @register('log-pane.scroll-to-top', key_bindings)
        def _scroll_to_top(_event: KeyPressEvent) -> None:
            """Scroll to top."""
            self.log_view.scroll_to_top()

        @register('log-pane.scroll-to-bottom', key_bindings)
        def _scroll_to_bottom(_event: KeyPressEvent) -> None:
            """Scroll to bottom."""
            self.log_view.scroll_to_bottom()

        @register('log-pane.toggle-follow', key_bindings)
        def _toggle_follow(_event: KeyPressEvent) -> None:
            """Toggle log line following."""
            self.log_pane.toggle_follow()

        @register('log-pane.toggle-web-browser', key_bindings)
        def _toggle_browser(_event: KeyPressEvent) -> None:
            """View logs in browser."""
            self.log_pane.toggle_websocket_server()

        @register('log-pane.move-cursor-up', key_bindings)
        def _up(_event: KeyPressEvent) -> None:
            """Move cursor up."""
            self.log_view.scroll_up()

        @register('log-pane.move-cursor-down', key_bindings)
        def _down(_event: KeyPressEvent) -> None:
            """Move cursor down."""
            self.log_view.scroll_down()

        @register('log-pane.visual-select-up', key_bindings)
        def _visual_select_up(_event: KeyPressEvent) -> None:
            """Select previous log line."""
            self.log_view.visual_select_up()

        @register('log-pane.visual-select-down', key_bindings)
        def _visual_select_down(_event: KeyPressEvent) -> None:
            """Select next log line."""
            self.log_view.visual_select_down()

        @register('log-pane.visual-select-all', key_bindings)
        def _select_all_logs(_event: KeyPressEvent) -> None:
            """Select all log lines."""
            self.log_pane.log_view.visual_select_all()

        @register('log-pane.scroll-page-up', key_bindings)
        def _pageup(_event: KeyPressEvent) -> None:
            """Scroll the logs up by one page."""
            self.log_view.scroll_up_one_page()

        @register('log-pane.scroll-page-down', key_bindings)
        def _pagedown(_event: KeyPressEvent) -> None:
            """Scroll the logs down by one page."""
            self.log_view.scroll_down_one_page()

        @register('log-pane.save-copy', key_bindings)
        def _start_saveas(_event: KeyPressEvent) -> None:
            """Save logs to a file."""
            self.log_pane.start_saveas()

        @register('log-pane.search', key_bindings)
        def _start_search(_event: KeyPressEvent) -> None:
            """Start searching."""
            self.log_pane.start_search()

        @register('log-pane.search-next-match', key_bindings)
        def _next_search(_event: KeyPressEvent) -> None:
            """Next search match."""
            self.log_view.search_forwards()

        @register('log-pane.search-previous-match', key_bindings)
        def _previous_search(_event: KeyPressEvent) -> None:
            """Previous search match."""
            self.log_view.search_backwards()

        @register('log-pane.deselect-cancel-search', key_bindings)
        def _clear_search_and_selection(_event: KeyPressEvent) -> None:
            """Clear selection or search."""
            if self.log_pane.log_view.visual_select_mode:
                self.log_pane.log_view.clear_visual_selection()
            elif self.log_pane.search_bar_active:
                self.log_pane.search_toolbar.cancel_search()

        @register('log-pane.search-apply-filter', key_bindings)
        def _apply_filter(_event: KeyPressEvent) -> None:
            """Apply current search as a filter."""
            self.log_pane.search_toolbar.close_search_bar()
            self.log_view.apply_filter()

        @register('log-pane.clear-filters', key_bindings)
        def _clear_filter(_event: KeyPressEvent) -> None:
            """Reset / erase active filters."""
            self.log_view.clear_filters()

        self.key_bindings: KeyBindingsBase = key_bindings

    def is_focusable(self) -> bool:
        return True

    def get_key_bindings(self) -> Optional[KeyBindingsBase]:
        return self.key_bindings

    def preferred_width(self, max_available_width: int) -> int:
        """Return the width of the longest line."""
        line_lengths = [len(l) for l in self.lines]
        return max(line_lengths)

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix,
    ) -> Optional[int]:
        """Return the preferred height for the log lines."""
        content = self.create_content(width, None)
        return content.line_count

    def create_content(self, width: int, height: Optional[int]) -> UIContent:
        # Update lines to render
        self.lines = self.log_view.render_content()

        # Create a UIContent instance if none exists
        if self.uicontent is None:
            self.uicontent = UIContent(
                get_line=lambda i: self.lines[i],
                line_count=len(self.lines),
                show_cursor=False,
            )

        # Update line_count
        self.uicontent.line_count = len(self.lines)

        return self.uicontent

    def mouse_handler(self, mouse_event: MouseEvent):
        """Mouse handler for this control."""
        mouse_position = mouse_event.position

        # Left mouse button release should:
        # 1. check if a mouse drag just completed.
        # 2. If not in focus, switch focus to this log pane
        #    If in focus, move the cursor to that position.
        if (
            mouse_event.event_type == MouseEventType.MOUSE_UP
            and mouse_event.button == MouseButton.LEFT
        ):
            # If a drag was in progress and this is the first mouse release
            # press, set the stop flag.
            if (
                self.visual_select_mode_drag_start
                and not self.visual_select_mode_drag_stop
            ):
                self.visual_select_mode_drag_stop = True

            if not has_focus(self)():
                # Focus the save as dialog if open.
                if self.log_pane.saveas_dialog_active:
                    get_app().layout.focus(self.log_pane.saveas_dialog)
                # Focus the search bar if open.
                elif self.log_pane.search_bar_active:
                    get_app().layout.focus(self.log_pane.search_toolbar)
                # Otherwise, focus on the log pane content.
                else:
                    get_app().layout.focus(self)
                # Mouse event handled, return None.
                return None

            # Log pane in focus already, move the cursor to the position of the
            # mouse click.
            self.log_pane.log_view.scroll_to_position(mouse_position)
            # Mouse event handled, return None.
            return None

        # Mouse drag with left button should start selecting lines.
        # The log pane does not need to be in focus to start this.
        if (
            mouse_event.event_type == MouseEventType.MOUSE_MOVE
            and mouse_event.button == MouseButton.LEFT
        ):
            # If a previous mouse drag was completed, clear the selection.
            if (
                self.visual_select_mode_drag_start
                and self.visual_select_mode_drag_stop
            ):
                self.log_pane.log_view.clear_visual_selection()
            # Drag select in progress, set flags accordingly.
            self.visual_select_mode_drag_start = True
            self.visual_select_mode_drag_stop = False

            self.log_pane.log_view.visual_select_line(mouse_position)
            # Mouse event handled, return None.
            return None

        # Mouse wheel events should move the cursor +/- some amount of lines
        # even if this pane is not in focus.
        if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            self.log_pane.log_view.scroll_down(lines=_LOG_OUTPUT_SCROLL_AMOUNT)
            # Mouse event handled, return None.
            return None

        if mouse_event.event_type == MouseEventType.SCROLL_UP:
            self.log_pane.log_view.scroll_up(lines=_LOG_OUTPUT_SCROLL_AMOUNT)
            # Mouse event handled, return None.
            return None

        # Mouse event not handled, return NotImplemented.
        return NotImplemented


class LogPaneWebsocketDialog(ConditionalContainer):
    """Dialog box for showing the websocket URL."""

    # Height of the dialog box contens in lines of text.
    DIALOG_HEIGHT = 2

    def __init__(self, log_pane: 'LogPane'):
        self.log_pane = log_pane

        self._last_action_message: str = ''
        self._last_action_time: float = 0

        info_bar_control = FormattedTextControl(self.get_info_fragments)
        info_bar_window = Window(
            content=info_bar_control,
            height=1,
            align=WindowAlign.LEFT,
            dont_extend_width=False,
        )

        message_bar_control = FormattedTextControl(self.get_message_fragments)
        message_bar_window = Window(
            content=message_bar_control,
            height=1,
            align=WindowAlign.RIGHT,
            dont_extend_width=False,
        )

        action_bar_control = FormattedTextControl(self.get_action_fragments)
        action_bar_window = Window(
            content=action_bar_control,
            height=1,
            align=WindowAlign.RIGHT,
            dont_extend_width=True,
        )

        super().__init__(
            create_border(
                HSplit(
                    [
                        info_bar_window,
                        VSplit([message_bar_window, action_bar_window]),
                    ],
                    height=LogPaneWebsocketDialog.DIALOG_HEIGHT,
                    style='class:saveas-dialog',
                ),
                content_height=LogPaneWebsocketDialog.DIALOG_HEIGHT,
                title='Websocket Log Server',
                border_style='class:saveas-dialog-border',
                left_margin_columns=1,
            ),
            filter=Condition(lambda: self.log_pane.websocket_dialog_active),
        )

    def focus_self(self) -> None:
        # Nothing in this dialog can be focused, focus on the parent log_pane
        # instead.
        self.log_pane.application.focus_on_container(self.log_pane)

    def close_dialog(self) -> None:
        """Close this dialog."""
        self.log_pane.toggle_websocket_server()
        self.log_pane.websocket_dialog_active = False
        self.log_pane.application.focus_on_container(self.log_pane)
        self.log_pane.redraw_ui()

    def _set_action_message(self, text: str) -> None:
        self._last_action_time = time.time()
        self._last_action_message = text

    def copy_url_to_clipboard(self) -> None:
        self.log_pane.application.application.clipboard.set_text(
            self.log_pane.log_view.get_web_socket_url()
        )
        self._set_action_message('Copied!')

    def get_message_fragments(self):
        """Return FormattedText with the last action message."""
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        # Separator should have the focus mouse handler so clicking on any
        # whitespace focuses the input field.
        separator_text = ('', '  ', focus)

        if self._last_action_time + 10 > time.time():
            return [
                ('class:theme-fg-yellow', self._last_action_message, focus),
                separator_text,
            ]
        return [separator_text]

    def get_info_fragments(self):
        """Return FormattedText with current URL info."""
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        # Separator should have the focus mouse handler so clicking on any
        # whitespace focuses the input field.
        separator_text = ('', '  ', focus)

        fragments = [
            ('class:saveas-dialog-setting', 'URL:  ', focus),
            (
                'class:saveas-dialog-title',
                self.log_pane.log_view.get_web_socket_url(),
                focus,
            ),
            separator_text,
        ]
        return fragments

    def get_action_fragments(self):
        """Return FormattedText with the action buttons."""
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        cancel = functools.partial(mouse_handlers.on_click, self.close_dialog)
        copy = functools.partial(
            mouse_handlers.on_click,
            self.copy_url_to_clipboard,
        )

        # Separator should have the focus mouse handler so clicking on any
        # whitespace focuses the input field.
        separator_text = ('', '  ', focus)

        # Default button style
        button_style = 'class:toolbar-button-inactive'

        fragments = []

        # Action buttons
        fragments.extend(
            to_keybind_indicator(
                key=None,
                description='Stop',
                mouse_handler=cancel,
                base_style=button_style,
            )
        )

        fragments.append(separator_text)
        fragments.extend(
            to_keybind_indicator(
                key=None,
                description='Copy to Clipboard',
                mouse_handler=copy,
                base_style=button_style,
            )
        )

        # One space separator
        fragments.append(('', ' ', focus))

        return fragments


class LogPane(WindowPane):
    """LogPane class."""

    # pylint: disable=too-many-instance-attributes,too-many-public-methods

    def __init__(
        self,
        application: Any,
        pane_title: str = 'Logs',
        log_store: Optional[LogStore] = None,
    ):
        super().__init__(application, pane_title)

        # TODO(tonymd): Read these settings from a project (or user) config.
        self.wrap_lines = False
        self._table_view = True
        self.is_a_duplicate = False

        # Create the log container which stores and handles incoming logs.
        self.log_view: LogView = LogView(
            self, self.application, log_store=log_store
        )

        # Log pane size variables. These are updated just befor rendering the
        # pane by the LogLineHSplit class.
        self.current_log_pane_width = 0
        self.current_log_pane_height = 0
        self.last_log_pane_width = None
        self.last_log_pane_height = None

        # Search tracking
        self.search_bar_active = False
        self.search_toolbar = SearchToolbar(self)
        self.filter_toolbar = FilterToolbar(self)

        self.saveas_dialog = LogPaneSaveAsDialog(self)
        self.saveas_dialog_active = False
        self.visual_selection_dialog = LogPaneSelectionDialog(self)

        self.websocket_dialog = LogPaneWebsocketDialog(self)
        self.websocket_dialog_active = False

        # Table header bar, only shown if table view is active.
        self.table_header_toolbar = TableToolbar(self)

        # Create the bottom toolbar for the whole log pane.
        self.bottom_toolbar = WindowPaneToolbar(self)
        self.bottom_toolbar.add_button(
            ToolbarButton('/', 'Search', self.start_search)
        )
        self.bottom_toolbar.add_button(
            ToolbarButton('Ctrl-o', 'Save', self.start_saveas)
        )
        self.bottom_toolbar.add_button(
            ToolbarButton(
                'f',
                'Follow',
                self.toggle_follow,
                is_checkbox=True,
                checked=lambda: self.log_view.follow,
            )
        )
        self.bottom_toolbar.add_button(
            ToolbarButton(
                't',
                'Table',
                self.toggle_table_view,
                is_checkbox=True,
                checked=lambda: self.table_view,
            )
        )
        self.bottom_toolbar.add_button(
            ToolbarButton(
                'w',
                'Wrap',
                self.toggle_wrap_lines,
                is_checkbox=True,
                checked=lambda: self.wrap_lines,
            )
        )
        self.bottom_toolbar.add_button(
            ToolbarButton('C', 'Clear', self.clear_history)
        )

        self.bottom_toolbar.add_button(
            ToolbarButton(
                'Shift-o',
                'Open in browser',
                self.toggle_websocket_server,
                is_checkbox=True,
                checked=lambda: self.log_view.websocket_running,
            )
        )

        self.log_content_control = LogContentControl(self)

        self.log_display_window = Window(
            content=self.log_content_control,
            # Scrolling is handled by LogScreen
            allow_scroll_beyond_bottom=False,
            # Line wrapping is handled by LogScreen
            wrap_lines=False,
            # Selected line highlighting is handled by LogScreen
            cursorline=False,
            # Don't make the window taller to fill the parent split container.
            # Window should match the height of the log line content. This will
            # also allow the parent HSplit to justify the content to the bottom
            dont_extend_height=True,
            # Window width should be extended to make backround highlighting
            # extend to the end of the container. Otherwise backround colors
            # will only appear until the end of the log line.
            dont_extend_width=False,
            # Needed for log lines ANSI sequences that don't specify foreground
            # or background colors.
            style=functools.partial(get_pane_style, self),
        )

        # Root level container
        self.container = ConditionalContainer(
            FloatContainer(
                # Horizonal split containing the log lines and the toolbar.
                WindowPaneHSplit(
                    self,  # LogPane reference
                    [
                        self.table_header_toolbar,
                        self.log_display_window,
                        self.filter_toolbar,
                        self.search_toolbar,
                        self.bottom_toolbar,
                    ],
                    # Align content with the bottom of the container.
                    align=VerticalAlign.BOTTOM,
                    height=lambda: self.height,
                    width=lambda: self.width,
                    style=functools.partial(get_pane_style, self),
                ),
                floats=[
                    Float(top=0, right=0, height=1, content=LineInfoBar(self)),
                    Float(
                        top=0,
                        right=0,
                        height=LogPaneSelectionDialog.DIALOG_HEIGHT,
                        content=self.visual_selection_dialog,
                    ),
                    Float(
                        top=3,
                        left=2,
                        right=2,
                        height=LogPaneSaveAsDialog.DIALOG_HEIGHT + 2,
                        content=self.saveas_dialog,
                    ),
                    Float(
                        top=1,
                        left=2,
                        right=2,
                        height=LogPaneWebsocketDialog.DIALOG_HEIGHT + 2,
                        content=self.websocket_dialog,
                    ),
                ],
            ),
            filter=Condition(lambda: self.show_pane),
        )

    @property
    def table_view(self):
        if self.log_view.websocket_running:
            return False
        return self._table_view

    @table_view.setter
    def table_view(self, table_view):
        self._table_view = table_view

    def menu_title(self):
        """Return the title to display in the Window menu."""
        title = self.pane_title()

        # List active filters
        if self.log_view.filtering_on:
            title += ' (FILTERS: '
            title += ' '.join(
                [
                    log_filter.pattern()
                    for log_filter in self.log_view.filters.values()
                ]
            )
            title += ')'
        return title

    def append_pane_subtitle(self, text):
        if not self._pane_subtitle:
            self._pane_subtitle = text
        else:
            self._pane_subtitle = self._pane_subtitle + ', ' + text

    def pane_subtitle(self) -> str:
        if not self._pane_subtitle:
            return ', '.join(self.log_view.log_store.channel_counts.keys())
        logger_names = self._pane_subtitle.split(', ')
        additional_text = ''
        if len(logger_names) > 1:
            additional_text = ' + {} more'.format(len(logger_names))

        return logger_names[0] + additional_text

    def start_search(self):
        """Show the search bar to begin a search."""
        if self.log_view.websocket_running:
            return
        # Show the search bar
        self.search_bar_active = True
        # Focus on the search bar
        self.application.focus_on_container(self.search_toolbar)

    def start_saveas(self, **export_kwargs) -> bool:
        """Show the saveas bar to begin saving logs to a file."""
        # Show the search bar
        self.saveas_dialog_active = True
        # Set export options if any
        self.saveas_dialog.set_export_options(**export_kwargs)
        # Focus on the search bar
        self.application.focus_on_container(self.saveas_dialog)
        return True

    def pane_resized(self) -> bool:
        """Return True if the current window size has changed."""
        return (
            self.last_log_pane_width != self.current_log_pane_width
            or self.last_log_pane_height != self.current_log_pane_height
        )

    def update_pane_size(self, width, height):
        """Save width and height of the log pane for the current UI render
        pass."""
        if width:
            self.last_log_pane_width = self.current_log_pane_width
            self.current_log_pane_width = width
        if height:
            # Subtract the height of the bottom toolbar
            height -= WindowPaneToolbar.TOOLBAR_HEIGHT
            if self._table_view:
                height -= TableToolbar.TOOLBAR_HEIGHT
            if self.search_bar_active:
                height -= SearchToolbar.TOOLBAR_HEIGHT
            if self.log_view.filtering_on:
                height -= FilterToolbar.TOOLBAR_HEIGHT
            self.last_log_pane_height = self.current_log_pane_height
            self.current_log_pane_height = height

    def toggle_table_view(self):
        """Enable or disable table view."""
        self._table_view = not self._table_view
        self.log_view.view_mode_changed()
        self.redraw_ui()

    def toggle_wrap_lines(self):
        """Enable or disable line wraping/truncation."""
        self.wrap_lines = not self.wrap_lines
        self.log_view.view_mode_changed()
        self.redraw_ui()

    def toggle_follow(self):
        """Enable or disable following log lines."""
        self.log_view.toggle_follow()
        self.redraw_ui()

    def clear_history(self):
        """Erase stored log lines."""
        self.log_view.clear_scrollback()
        self.redraw_ui()

    def toggle_websocket_server(self):
        """Start or stop websocket server to send logs."""
        if self.log_view.websocket_running:
            self.log_view.stop_websocket_thread()
            self.websocket_dialog_active = False
        else:
            self.search_toolbar.close_search_bar()
            self.log_view.start_websocket_thread()
            self.application.start_http_server()
            self.saveas_dialog_active = False
            self.websocket_dialog_active = True

    def get_all_key_bindings(self) -> List:
        """Return all keybinds for this pane."""
        # Return log content control keybindings
        return [self.log_content_control.get_key_bindings()]

    def get_window_menu_options(
        self,
    ) -> List[Tuple[str, Union[Callable, None]]]:
        """Return all menu options for the log pane."""

        options = [
            # Menu separator
            ('-', None),
            (
                'Save/Export a copy',
                self.start_saveas,
            ),
            ('-', None),
            (
                '{check} Line wrapping'.format(
                    check=to_checkbox_text(self.wrap_lines, end='')
                ),
                self.toggle_wrap_lines,
            ),
            (
                '{check} Table view'.format(
                    check=to_checkbox_text(self._table_view, end='')
                ),
                self.toggle_table_view,
            ),
            (
                '{check} Follow'.format(
                    check=to_checkbox_text(self.log_view.follow, end='')
                ),
                self.toggle_follow,
            ),
            (
                '{check} Open in web browser'.format(
                    check=to_checkbox_text(
                        self.log_view.websocket_running, end=''
                    )
                ),
                self.toggle_websocket_server,
            ),
            # Menu separator
            ('-', None),
            (
                'Clear history',
                self.clear_history,
            ),
            (
                'Duplicate pane',
                self.duplicate,
            ),
        ]
        if self.is_a_duplicate:
            options += [
                (
                    'Remove/Delete pane',
                    functools.partial(
                        self.application.window_manager.remove_pane, self
                    ),
                )
            ]

        # Search / Filter section
        options += [
            # Menu separator
            ('-', None),
            (
                'Hide search highlighting',
                self.log_view.disable_search_highlighting,
            ),
            (
                'Create filter from search results',
                self.log_view.apply_filter,
            ),
            (
                'Clear/Reset active filters',
                self.log_view.clear_filters,
            ),
        ]

        return options

    def apply_filters_from_config(self, window_options) -> None:
        if 'filters' not in window_options:
            return

        for field, criteria in window_options['filters'].items():
            for matcher_name, search_string in criteria.items():
                inverted = matcher_name.endswith('-inverted')
                matcher_name = re.sub(r'-inverted$', '', matcher_name)
                if field == 'all':
                    field = None
                if self.log_view.new_search(
                    search_string,
                    invert=inverted,
                    field=field,
                    search_matcher=matcher_name,
                    interactive=False,
                ):
                    self.log_view.install_new_filter()

        # Trigger any existing log messages to be added to the view.
        self.log_view.new_logs_arrived()

    def create_duplicate(self) -> 'LogPane':
        """Create a duplicate of this LogView."""
        new_pane = LogPane(self.application, pane_title=self.pane_title())
        # Set the log_store
        log_store = self.log_view.log_store
        new_pane.log_view.log_store = log_store
        # Register the duplicate pane as a viewer
        log_store.register_viewer(new_pane.log_view)

        # Set any existing search state.
        new_pane.log_view.search_text = self.log_view.search_text
        new_pane.log_view.search_filter = self.log_view.search_filter
        new_pane.log_view.search_matcher = self.log_view.search_matcher
        new_pane.log_view.search_highlight = self.log_view.search_highlight

        # Mark new pane as a duplicate so it can be deleted.
        new_pane.is_a_duplicate = True
        return new_pane

    def duplicate(self) -> None:
        new_pane = self.create_duplicate()
        # Add the new pane.
        self.application.window_manager.add_pane(new_pane)

    def add_log_handler(
        self,
        logger: Union[str, logging.Logger],
        level_name: Optional[str] = None,
    ) -> None:
        """Add a log handlers to this LogPane."""

        if isinstance(logger, logging.Logger):
            logger_instance = logger
        elif isinstance(logger, str):
            logger_instance = logging.getLogger(logger)

        if level_name:
            if not hasattr(logging, level_name):
                raise Exception(f'Unknown log level: {level_name}')
            logger_instance.level = getattr(logging, level_name, logging.INFO)
        logger_instance.addHandler(self.log_view.log_store)  # type: ignore
        self.append_pane_subtitle(logger_instance.name)  # type: ignore



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Sphinx directives for Pigweed module metadata"""

from typing import List

import docutils

# pylint: disable=consider-using-from-import
import docutils.parsers.rst.directives as directives  # type: ignore

# pylint: enable=consider-using-from-import
from sphinx.application import Sphinx as SphinxApplication
from sphinx.util.docutils import SphinxDirective
from sphinx_design.badges_buttons import ButtonRefDirective  # type: ignore
from sphinx_design.cards import CardDirective  # type: ignore


def status_choice(arg):
    return directives.choice(arg, ('experimental', 'unstable', 'stable'))


def status_badge(module_status: str) -> str:
    role = ':bdg-primary:'
    return role + f'`{module_status.title()}`'


def cs_url(module_name: str):
    return f'https://cs.opensource.google/pigweed/pigweed/+/main:{module_name}/'


def concat_tags(*tag_lists: List[str]):
    all_tags = tag_lists[0]

    for tag_list in tag_lists[1:]:
        if len(tag_list) > 0:
            all_tags.append(':octicon:`dot-fill`')
            all_tags.extend(tag_list)

    return all_tags


class PigweedModuleDirective(SphinxDirective):
    """Directive registering module metadata, rendering title & info card."""

    required_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        'name': directives.unchanged_required,
        'tagline': directives.unchanged_required,
        'status': status_choice,
        'is-deprecated': directives.flag,
        'languages': directives.unchanged,
        'code-size-impact': directives.unchanged,
        'facade': directives.unchanged,
        'get-started': directives.unchanged_required,
        'tutorials': directives.unchanged,
        'guides': directives.unchanged,
        'concepts': directives.unchanged,
        'design': directives.unchanged_required,
        'api': directives.unchanged,
    }

    def try_get_option(self, option: str):
        try:
            return self.options[option]
        except KeyError:
            raise self.error(f' :{option}: option is required')

    def maybe_get_option(self, option: str):
        try:
            return self.options[option]
        except KeyError:
            return None

    def create_section_button(self, title: str, ref: str):
        node = docutils.nodes.list_item(classes=['pw-module-section-button'])
        node += ButtonRefDirective(
            name='',
            arguments=[ref],
            options={'color': 'primary'},
            content=[title],
            lineno=0,
            content_offset=0,
            block_text='',
            state=self.state,
            state_machine=self.state_machine,
        ).run()

        return node

    def register_metadata(self):
        module_name = self.try_get_option('name')

        if 'facade' in self.options:
            facade = self.options['facade']

            # Initialize the module relationship dict if needed
            if not hasattr(self.env, 'pw_module_relationships'):
                self.env.pw_module_relationships = {}

            # Initialize the backend list for this facade if needed
            if facade not in self.env.pw_module_relationships:
                self.env.pw_module_relationships[facade] = []

            # Add this module as a backend of the provided facade
            self.env.pw_module_relationships[facade].append(module_name)

        if 'is-deprecated' in self.options:
            # Initialize the deprecated modules list if needed
            if not hasattr(self.env, 'pw_modules_deprecated'):
                self.env.pw_modules_deprecated = []

            self.env.pw_modules_deprecated.append(module_name)

    def run(self):
        tagline = docutils.nodes.paragraph(
            text=self.try_get_option('tagline'),
            classes=['section-subtitle'],
        )

        status_tags: List[str] = [
            status_badge(self.try_get_option('status')),
        ]

        if 'is-deprecated' in self.options:
            status_tags.append(':bdg-danger:`Deprecated`')

        language_tags = []

        if 'languages' in self.options:
            languages = self.options['languages'].split(',')

            if len(languages) > 0:
                for language in languages:
                    language_tags.append(f':bdg-info:`{language.strip()}`')

        code_size_impact = []

        if code_size_text := self.maybe_get_option('code-size-impact'):
            code_size_impact.append(f'**Code Size Impact:** {code_size_text}')

        # Move the directive content into a section that we can render wherever
        # we want.
        content = docutils.nodes.paragraph()
        self.state.nested_parse(self.content, 0, content)

        # The card inherits its content from this node's content, which we've
        # already pulled out. So we can replace this node's content with the
        # content we need in the card.
        self.content = docutils.statemachine.StringList(
            concat_tags(status_tags, language_tags, code_size_impact)
        )

        card = CardDirective.create_card(
            inst=self,
            arguments=[],
            options={},
        )

        # Create the top-level section buttons.
        section_buttons = docutils.nodes.bullet_list(
            classes=['pw-module-section-buttons']
        )

        # This is the pattern for required sections.
        section_buttons += self.create_section_button(
            'Get Started', self.try_get_option('get-started')
        )

        # This is the pattern for optional sections.
        if (tutorials_ref := self.maybe_get_option('tutorials')) is not None:
            section_buttons += self.create_section_button(
                'Tutorials', tutorials_ref
            )

        if (guides_ref := self.maybe_get_option('guides')) is not None:
            section_buttons += self.create_section_button('Guides', guides_ref)

        if (concepts_ref := self.maybe_get_option('concepts')) is not None:
            section_buttons += self.create_section_button(
                'Concepts', concepts_ref
            )

        section_buttons += self.create_section_button(
            'Design', self.try_get_option('design')
        )

        if (api_ref := self.maybe_get_option('api')) is not None:
            section_buttons += self.create_section_button(
                'API Reference', api_ref
            )

        return [tagline, section_buttons, content, card]


def build_backend_lists(app, _doctree, _fromdocname):
    env = app.builder.env

    if not hasattr(env, 'pw_module_relationships'):
        env.pw_module_relationships = {}


def setup(app: SphinxApplication):
    app.add_directive('pigweed-module', PigweedModuleDirective)

    # At this event, the documents and metadata have been generated, and now we
    # can modify the doctree to reflect the metadata.
    app.connect('doctree-resolved', build_backend_lists)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



#!/usr/bin/env python3
# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Command line interface for stm32cube_builder."""

import argparse
import pathlib
import sys

try:
    from pw_stm32cube_build import (
        find_files,
        gen_file_list,
        icf_to_ld,
        inject_init,
    )
except ImportError:
    # Load from this directory if pw_stm32cube_build is not available.
    import find_files  # type: ignore
    import gen_file_list  # type: ignore
    import icf_to_ld  # type: ignore
    import inject_init  # type: ignore


def _parse_args() -> argparse.Namespace:
    """Setup argparse and parse command line args."""
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(
        dest='command', metavar='<command>', required=True
    )

    gen_file_list_parser = subparsers.add_parser(
        'gen_file_list', help='generate files.txt for stm32cube directory'
    )
    gen_file_list_parser.add_argument('stm32cube_dir', type=pathlib.Path)

    find_files_parser = subparsers.add_parser(
        'find_files', help='find files in stm32cube directory'
    )
    find_files_parser.add_argument('stm32cube_dir', type=pathlib.Path)
    find_files_parser.add_argument('product_str')
    find_files_parser.add_argument('--init', default=False, action='store_true')

    icf_to_ld_parser = subparsers.add_parser(
        'icf_to_ld', help='convert stm32cube .icf linker files to .ld'
    )
    icf_to_ld_parser.add_argument('icf_path', type=pathlib.Path)
    icf_to_ld_parser.add_argument(
        '--ld-path', nargs=1, default=None, type=pathlib.Path
    )

    inject_init_parser = subparsers.add_parser(
        'inject_init', help='inject `pw_stm32cube_Init()` into startup_*.s'
    )
    inject_init_parser.add_argument('in_startup_path', type=pathlib.Path)
    inject_init_parser.add_argument(
        '--out-startup-path', nargs=1, default=None, type=pathlib.Path
    )

    return parser.parse_args()


def main():
    """Main command line function."""
    args = _parse_args()

    if args.command == 'gen_file_list':
        gen_file_list.gen_file_list(args.stm32cube_dir)
    elif args.command == 'find_files':
        find_files.find_files(args.stm32cube_dir, args.product_str, args.init)
    elif args.command == 'icf_to_ld':
        icf_to_ld.icf_to_ld(
            args.icf_path, args.ld_path[0] if args.ld_path else None
        )
    elif args.command == 'inject_init':
        inject_init.inject_init(
            args.in_startup_path,
            args.out_startup_path[0] if args.out_startup_path else None,
        )

    sys.exit(0)


if __name__ == '__main__':
    main()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Configure Visual Studio Code (VSC) for Pigweed projects.

VSC recognizes three sources of configurable settings:

1. Project settings, stored in {project root}/.vscode/settings.json
2. Workspace settings, stored in (workspace root)/.vscode/settings.json;
   a workspace is a collection of projects/repositories that are worked on
   together in a single VSC instance
3. The user's personal settings, which are stored somewhere in the user's home
   directory, and are applied to all instances of VSC

This provides three levels of settings cascading:

    Workspace <- Project <- User

... where the arrow indicates the ability to override.

Out of these three, only project settings are useful to Pigweed projects. User
settings are essentially global and outside the scope of Pigweed. Workspaces are
seldom used and don't work well with the Pigweed directory structure.

Nonetheless, we want a three-tiered settings structure for Pigweed projects too:

A. Default settings provided by Pigweed, configuring VSC to use IDE features
B. Project-level overrides that downstream projects may define
C. User-level overrides that individual users may define

We accomplish all of that with only the project settings described in #1 above.

Default settings are defined in this module. Project settings can be defined in
.vscode/pw_project_settings.json and should be checked into the repository. User
settings can be defined in .vscode/pw_user_settings.json and should not be
checked into the repository. None of these settings have any effect until they
are merged into VSC's settings (.vscode/settings.json) via the functions in this
module. Those resulting settings are system-specific and should also not be
checked into the repository.

We provide the same structure to both tasks and extensions as well. Defaults
are provided by Pigweed, can be augmented or overridden at the project level
with .vscode/pw_project_tasks.json and .vscode/pw_project_extensions.json,
can be augmented or overridden by an individual developer with
.vscode/pw_user_tasks.json and .vscode/pw_user.extensions.json, and none of
this takes effect until they are merged into VSC's active settings files
(.vscode/tasks.json and .vscode/extensions.json) by running the appropriate
command.
"""

# TODO(chadnorvell): Import collections.OrderedDict when we don't need to
# support Python 3.8 anymore.
from enum import Enum
import json
import os
from pathlib import Path
import platform
from typing import Any, Dict, List, OrderedDict

from pw_ide.activate import BashShellModifier
from pw_ide.cpp import ClangdSettings

from pw_ide.editors import (
    EditorSettingsDict,
    EditorSettingsManager,
    EditorSettingsTypesWithDefaults,
    Json5FileFormat,
)

from pw_ide.python import PythonPaths
from pw_ide.settings import PigweedIdeSettings


def _vsc_os(system: str = platform.system()):
    """Return the OS tag that VSC expects."""
    if system == 'Darwin':
        return 'osx'

    return system.lower()


def _activated_env() -> OrderedDict[str, Any]:
    """Return the environment diff needed to provide Pigweed activation.

    The integrated terminal will already include the user's default environment
    (e.g. from their shell init scripts). This provides the modifications to
    the environment needed for Pigweed activation.
    """
    # Not all environments have an actions.json, which this ultimately relies
    # on (e.g. tests in CI). No problem, just return an empty dict instead.
    try:
        env = (
            BashShellModifier(env_only=True, path_var='${env:PATH}')
            .modify_env()
            .env_mod
        )
    except (FileNotFoundError, json.JSONDecodeError):
        env = dict()

    return OrderedDict(env)


def _local_terminal_integrated_env() -> Dict[str, Any]:
    """VSC setting to activate the integrated terminal."""
    return {f'terminal.integrated.env.{_vsc_os()}': _activated_env()}


def _local_clangd_settings(ide_settings: PigweedIdeSettings) -> Dict[str, Any]:
    """VSC settings for running clangd with Pigweed paths."""
    clangd_settings = ClangdSettings(ide_settings)
    return {
        'clangd.path': str(clangd_settings.clangd_path),
        'clangd.arguments': clangd_settings.arguments,
    }


def _local_python_settings() -> Dict[str, Any]:
    """VSC settings for finding the Python virtualenv."""
    paths = PythonPaths()
    return {
        'python.defaultInterpreterPath': str(paths.interpreter),
        'python.formatting.yapfPath': str(paths.bin_dir / 'yapf'),
    }


# The order is preserved despite starting with a plain dict because in Python
# 3.6+, plain dicts are actually ordered as an implementation detail. This could
# break on interpreters other than CPython, or if the implementation changes in
# the future. However, for now, this is much more readable than the more robust
# alternative of instantiating with a list of tuples.
_DEFAULT_SETTINGS: EditorSettingsDict = OrderedDict(
    {
        "editor.detectIndentation": False,
        "editor.rulers": [80],
        "editor.tabSize": 2,
        "files.associations": OrderedDict({"*.inc": "cpp"}),
        "files.exclude": OrderedDict(
            {
                "**/*.egg-info": True,
                "**/.mypy_cache": True,
                "**/__pycache__": True,
                ".cache": True,
                ".cipd": True,
                ".environment": True,
                ".presubmit": True,
                ".pw_ide": True,
                ".pw_ide.user.yaml": True,
                "bazel-bin": True,
                "bazel-out": True,
                "bazel-pigweed": True,
                "bazel-testlogs": True,
                "build": True,
                "environment": True,
                "node_modules": True,
                "out": True,
            }
        ),
        "files.insertFinalNewline": True,
        "files.trimTrailingWhitespace": True,
        "search.useGlobalIgnoreFiles": True,
        "grunt.autoDetect": "off",
        "gulp.autoDetect": "off",
        "jake.autoDetect": "off",
        "npm.autoDetect": "off",
        "clangd.onConfigChanged": "restart",
        "C_Cpp.intelliSenseEngine": "Disabled",
        "[cpp]": OrderedDict(
            {"editor.defaultFormatter": "llvm-vs-code-extensions.vscode-clangd"}
        ),
        "python.formatting.provider": "yapf",
        "[python]": OrderedDict({"editor.tabSize": 4}),
        "typescript.tsc.autoDetect": "off",
        "[gn]": OrderedDict({"editor.defaultFormatter": "msedge-dev.gnls"}),
        "[proto3]": OrderedDict(
            {"editor.defaultFormatter": "zxh404.vscode-proto3"}
        ),
    }
)

# pylint: disable=line-too-long
_DEFAULT_TASKS: EditorSettingsDict = OrderedDict(
    {
        "version": "2.0.0",
        "tasks": [
            {
                "type": "shell",
                "label": "Pigweed IDE: Format",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw format --fix'",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: Presubmit",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw presubmit'",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed IDE: Set Python Virtual Environment",
                "command": "${command:python.setInterpreter}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed IDE: Restart Python Language Server",
                "command": "${command:python.analysis.restartLanguageServer}",
                "problemMatcher": [],
            },
            {
                "label": "Pigweed IDE: Restart C++ Language Server",
                "command": "${command:clangd.restart}",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: Process C++ Compilation Database from GN",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw ide cpp --gn --process out/compile_commands.json'",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: Setup",
                "command": "python3 ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw ide setup'",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: Current C++ Code Analysis Target",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw ide cpp'",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: List C++ Code Analysis Targets",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw ide cpp --list'",
                "problemMatcher": [],
            },
            {
                "type": "shell",
                "label": "Pigweed IDE: Set C++ Code Analysis Target",
                "command": "${workspaceFolder}/.pw_ide/python ${workspaceFolder}/pw_ide/py/pw_ide/activate.py -x 'pw ide cpp --set ${input:target}'",
                "problemMatcher": [],
            },
        ],
        "inputs": [
            {
                "id": "target",
                "type": "promptString",
                "description": "C++ code analysis target",
            }
        ],
    }
)
# pylint: enable=line-too-long

_DEFAULT_EXTENSIONS: EditorSettingsDict = OrderedDict(
    {
        "recommendations": [
            "llvm-vs-code-extensions.vscode-clangd",
            "ms-python.python",
            "npclaudiu.vscode-gn",
            "msedge-dev.gnls",
            "zxh404.vscode-proto3",
            "josetr.cmake-language-support-vscode",
            "swyddfa.esbonio",
        ],
        "unwantedRecommendations": [
            "ms-vscode.cpptools",
            "persidskiy.vscode-gnformat",
            "lextudio.restructuredtext",
            "trond-snekvik.simple-rst",
        ],
    }
)


def _default_settings(
    pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return OrderedDict(
        {
            **_DEFAULT_SETTINGS,
            **_local_terminal_integrated_env(),
            **_local_clangd_settings(pw_ide_settings),
            **_local_python_settings(),
        }
    )


def _default_tasks(_pw_ide_settings: PigweedIdeSettings) -> EditorSettingsDict:
    return _DEFAULT_TASKS


def _default_extensions(
    _pw_ide_settings: PigweedIdeSettings,
) -> EditorSettingsDict:
    return _DEFAULT_EXTENSIONS


DEFAULT_SETTINGS_PATH = Path(os.path.expandvars('$PW_PROJECT_ROOT')) / '.vscode'


class VscSettingsType(Enum):
    """Visual Studio Code settings files.

    VSC supports editor settings (``settings.json``), recommended
    extensions (``extensions.json``), and tasks (``tasks.json``).
    """

    SETTINGS = 'settings'
    TASKS = 'tasks'
    EXTENSIONS = 'extensions'

    @classmethod
    def all(cls) -> List['VscSettingsType']:
        return list(cls)


class VscSettingsManager(EditorSettingsManager[VscSettingsType]):
    """Manages all settings for Visual Studio Code."""

    default_settings_dir = DEFAULT_SETTINGS_PATH
    file_format = Json5FileFormat()

    types_with_defaults: EditorSettingsTypesWithDefaults = {
        VscSettingsType.SETTINGS: _default_settings,
        VscSettingsType.TASKS: _default_tasks,
        VscSettingsType.EXTENSIONS: _default_extensions,
    }



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Example text input-output Plugin."""

from typing import TYPE_CHECKING

from prompt_toolkit.document import Document
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.layout import Window
from prompt_toolkit.widgets import SearchToolbar, TextArea

from pw_console.widgets import ToolbarButton, WindowPane, WindowPaneToolbar

if TYPE_CHECKING:
    from pw_console.console_app import ConsoleApp


class CalcPane(WindowPane):
    """Example plugin that accepts text input and displays output.

    This plugin is similar to the full-screen calculator example provided in
    prompt_toolkit:
    https://github.com/prompt-toolkit/python-prompt-toolkit/blob/3.0.23/examples/full-screen/calculator.py

    It's a full window that can be moved around the user interface like other
    Pigweed Console window panes. An input prompt is displayed on the bottom of
    the window where the user can type in some math equation. When the enter key
    is pressed the input is processed and the result shown in the top half of
    the window.

    Both input and output fields are prompt_toolkit TextArea objects which can
    have their own options like syntax highlighting.
    """

    def __init__(self):
        # Call WindowPane.__init__ and set the title to 'Calculator'
        super().__init__(pane_title='Calculator')

        # Create a TextArea for the output-field
        # TextArea is a prompt_toolkit widget that can display editable text in
        # a buffer. See the prompt_toolkit docs for all possible options:
        # https://python-prompt-toolkit.readthedocs.io/en/latest/pages/reference.html#prompt_toolkit.widgets.TextArea
        self.output_field = TextArea(
            # Optional Styles to apply to this TextArea
            style='class:output-field',
            # Initial text to put into the buffer.
            text='Calculator Output',
            # Allow this buffer to be in focus. This lets you drag select text
            # contained inside, and edit the contents unless readonly.
            focusable=True,
            # Focus on mouse click.
            focus_on_click=True,
        )

        # This is the search toolbar and only appears if the user presses ctrl-r
        # to do reverse history search (similar to bash or zsh). Its used by the
        # input_field below.
        self.search_field = SearchToolbar()

        # Create a TextArea for the user input.
        self.input_field = TextArea(
            # The height is set to 1 line
            height=1,
            # Prompt string that appears before the cursor.
            prompt='>>> ',
            # Optional Styles to apply to this TextArea
            style='class:input-field',
            # We only allow one line input for this example but multiline is
            # supported by prompt_toolkit.
            multiline=False,
            wrap_lines=False,
            # Allow reverse history search
            search_field=self.search_field,
            # Allow this input to be focused.
            focusable=True,
            # Focus on mouse click.
            focus_on_click=True,
        )

        # The TextArea accept_handler function is called by prompt_toolkit (the
        # UI) when the user presses enter. Here we override it to our own accept
        # handler defined in this CalcPane class.
        self.input_field.accept_handler = self.accept_input

        # Create a toolbar for display at the bottom of this window. It will
        # show the window title and toolbar buttons.
        self.bottom_toolbar = WindowPaneToolbar(self)
        self.bottom_toolbar.add_button(
            ToolbarButton(
                key='Enter',  # Key binding for this function
                description='Run Calculation',  # Button name
                # Function to run when clicked.
                mouse_handler=self.run_calculation,
            )
        )
        self.bottom_toolbar.add_button(
            ToolbarButton(
                key='Ctrl-c',  # Key binding for this function
                description='Copy Output',  # Button name
                # Function to run when clicked.
                mouse_handler=self.copy_all_output,
            )
        )

        # self.container is the root container that contains objects to be
        # rendered in the UI, one on top of the other.
        self.container = self._create_pane_container(
            # Show the output_field on top
            self.output_field,
            # Draw a separator line with height=1
            Window(height=1, char='─', style='class:line'),
            # Show the input field just below that.
            self.input_field,
            # If ctrl-r reverse history is active, show the search box below the
            # input_field.
            self.search_field,
            # Lastly, show the toolbar.
            self.bottom_toolbar,
        )

    def pw_console_init(self, app: 'ConsoleApp') -> None:
        """Set the Pigweed Console application instance.

        This function is called after the Pigweed Console starts up and allows
        access to the user preferences. Prefs is required for creating new
        user-remappable keybinds."""
        self.application = app
        self.set_custom_keybinds()

    def set_custom_keybinds(self) -> None:
        # Fetch ConsoleApp preferences to load user keybindings
        prefs = self.application.prefs
        # Register a named keybind function that is user re-mappable
        prefs.register_named_key_function(
            'calc-pane.copy-selected-text',
            # default bindings
            ['c-c'],
        )

        # For setting additional keybindings to the output_field.
        key_bindings = KeyBindings()

        # Map the 'calc-pane.copy-selected-text' function keybind to the
        # _copy_all_output function below. This will set
        @prefs.register_keybinding('calc-pane.copy-selected-text', key_bindings)
        def _copy_all_output(_event: KeyPressEvent) -> None:
            """Copy selected text from the output buffer."""
            self.copy_selected_output()

        # Set the output_field controls key_bindings to the new bindings.
        self.output_field.control.key_bindings = key_bindings

    def run_calculation(self):
        """Trigger the input_field's accept_handler.

        This has the same effect as pressing enter in the input_field.
        """
        self.input_field.buffer.validate_and_handle()

    def accept_input(self, _buffer):
        """Function run when the user presses enter in the input_field.

        Takes a buffer argument that contains the user's input text.
        """
        # Evaluate the user's calculator expression as Python and format the
        # output result.
        try:
            output = "\n\nIn:  {}\nOut: {}".format(
                self.input_field.text,
                # NOTE: Don't use 'eval' in real code (this is just an example)
                eval(self.input_field.text),  # pylint: disable=eval-used
            )
        except BaseException as exception:  # pylint: disable=broad-except
            output = "\n\n{}".format(exception)

        # Append the new output result to the existing output_field contents.
        new_text = self.output_field.text + output

        # Update the output_field with the new contents and move the
        # cursor_position to the end.
        self.output_field.buffer.document = Document(
            text=new_text, cursor_position=len(new_text)
        )

    def copy_selected_output(self):
        """Copy highlighted text in the output_field to the system clipboard."""
        clipboard_data = self.output_field.buffer.copy_selection()
        self.application.application.clipboard.set_data(clipboard_data)

    def copy_all_output(self):
        """Copy all text in the output_field to the system clipboard."""
        self.application.application.clipboard.set_text(
            self.output_field.buffer.text
        )



#!/usr/bin/env python3
# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates flags needed for an ARM build using clang.

Using clang on Cortex-M cores isn't intuitive as the end-to-end experience isn't
quite completely in LLVM. LLVM doesn't yet provide compatible C runtime
libraries or C/C++ standard libraries. To work around this, this script pulls
the missing bits from an arm-none-eabi-gcc compiler on the system path. This
lets clang do the heavy lifting while only relying on some headers provided by
newlib/arm-none-eabi-gcc in addition to a small assortment of needed libraries.

To use this script, specify what flags you want from the script, and run with
the required architecture flags like you would with gcc:

  python -m pw_toolchain.clang_arm_toolchain --cflags -- -mthumb -mcpu=cortex-m3

The script will then print out the additional flags you need to pass to clang to
get a working build.
"""

import argparse
import sys
import os
import subprocess

from pathlib import Path
from typing import List, Dict, Tuple

_ARM_COMPILER_PREFIX = 'arm-none-eabi'
_ARM_COMPILER_NAME = _ARM_COMPILER_PREFIX + '-gcc'


def _parse_args() -> argparse.Namespace:
    """Parses arguments for this script, splitting out the command to run."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--gn-scope',
        action='store_true',
        help=(
            "Formats the output like a GN scope so it can be ingested by "
            "exec_script()"
        ),
    )
    parser.add_argument(
        '--cflags',
        action='store_true',
        help=('Include necessary C flags in the output'),
    )
    parser.add_argument(
        '--ldflags',
        action='store_true',
        help=('Include necessary linker flags in the output'),
    )
    parser.add_argument(
        'clang_flags',
        nargs=argparse.REMAINDER,
        help='Flags to pass to clang, which can affect library/include paths',
    )
    parsed_args = parser.parse_args()

    assert parsed_args.clang_flags[0] == '--', 'arguments not correctly split'
    parsed_args.clang_flags = parsed_args.clang_flags[1:]
    return parsed_args


def _compiler_info_command(print_command: str, cflags: List[str]) -> str:
    command = [_ARM_COMPILER_NAME]
    command.extend(cflags)
    command.append(print_command)
    result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    result.check_returncode()
    return result.stdout.decode().rstrip()


def get_gcc_lib_dir(cflags: List[str]) -> Path:
    return Path(
        _compiler_info_command('-print-libgcc-file-name', cflags)
    ).parent


def get_compiler_info(cflags: List[str]) -> Dict[str, str]:
    compiler_info: Dict[str, str] = {}
    compiler_info['gcc_libs_dir'] = os.path.relpath(
        str(get_gcc_lib_dir(cflags)), "."
    )
    compiler_info['sysroot'] = os.path.relpath(
        _compiler_info_command('-print-sysroot', cflags), "."
    )
    compiler_info['version'] = _compiler_info_command('-dumpversion', cflags)
    compiler_info['multi_dir'] = _compiler_info_command(
        '-print-multi-directory', cflags
    )
    return compiler_info


def get_cflags(compiler_info: Dict[str, str]):
    """TODO(amontanez): Add docstring."""
    # TODO(amontanez): Make newlib-nano optional.
    cflags = [
        # TODO(amontanez): For some reason, -stdlib++-isystem and
        # -isystem-after work, but emit unused argument errors. This is the only
        # way to let the build succeed.
        '-Qunused-arguments',
        # Disable all default libraries.
        "-nodefaultlibs",
        '--target=arm-none-eabi',
    ]

    # Add sysroot info.
    cflags.extend(
        (
            '--sysroot=' + compiler_info['sysroot'],
            '-isystem'
            + str(Path(compiler_info['sysroot']) / 'include' / 'newlib-nano'),
            # This must be included after Clang's builtin headers.
            '-isystem-after' + str(Path(compiler_info['sysroot']) / 'include'),
            '-stdlib++-isystem'
            + str(
                Path(compiler_info['sysroot'])
                / 'include'
                / 'c++'
                / compiler_info['version']
            ),
            '-isystem'
            + str(
                Path(compiler_info['sysroot'])
                / 'include'
                / 'c++'
                / compiler_info['version']
                / _ARM_COMPILER_PREFIX
                / compiler_info['multi_dir']
            ),
        )
    )

    return cflags


def get_crt_objs(compiler_info: Dict[str, str]) -> Tuple[str, ...]:
    return (
        str(Path(compiler_info['gcc_libs_dir']) / 'crtfastmath.o'),
        str(Path(compiler_info['gcc_libs_dir']) / 'crti.o'),
        str(Path(compiler_info['gcc_libs_dir']) / 'crtn.o'),
        str(
            Path(compiler_info['sysroot'])
            / 'lib'
            / compiler_info['multi_dir']
            / 'crt0.o'
        ),
    )


def get_ldflags(compiler_info: Dict[str, str]) -> List[str]:
    ldflags: List[str] = [
        # Add library search paths.
        '-L' + compiler_info['gcc_libs_dir'],
        '-L'
        + str(
            Path(compiler_info['sysroot']) / 'lib' / compiler_info['multi_dir']
        ),
        # Add libraries to link.
        '-lc_nano',
        '-lm',
        '-lgcc',
        '-lstdc++_nano',
    ]

    # Add C runtime object files.
    objs = get_crt_objs(compiler_info)
    ldflags.extend(objs)

    return ldflags


def main(
    cflags: bool,
    ldflags: bool,
    gn_scope: bool,
    clang_flags: List[str],
) -> int:
    """Script entry point."""
    compiler_info = get_compiler_info(clang_flags)
    if ldflags:
        ldflag_list = get_ldflags(compiler_info)

    if cflags:
        cflag_list = get_cflags(compiler_info)

    if not gn_scope:
        flags = []
        if cflags:
            flags.extend(cflag_list)
        if ldflags:
            flags.extend(ldflag_list)
        print(' '.join(flags))
        return 0

    if cflags:
        print('cflags = [')
        for flag in cflag_list:
            print(f'  "{flag}",')
        print(']')

    if ldflags:
        print('ldflags = [')
        for flag in ldflag_list:
            print(f'  "{flag}",')
        print(']')
    return 0


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Table view renderer for LogLines."""

import collections
import copy

from prompt_toolkit.formatted_text import StyleAndTextTuples

from pw_console.console_prefs import ConsolePrefs
from pw_console.log_line import LogLine
from pw_console.text_formatting import strip_ansi


class TableView:
    """Store column information and render logs into formatted tables."""

    # TODO(tonymd): Add a method to provide column formatters externally.
    # Should allow for string format, column color, and column ordering.
    FLOAT_FORMAT = '%.3f'
    INT_FORMAT = '%s'
    LAST_TABLE_COLUMN_NAMES = ['msg', 'message']

    def __init__(self, prefs: ConsolePrefs):
        self.set_prefs(prefs)
        self.column_widths: collections.OrderedDict = collections.OrderedDict()
        self._header_fragment_cache = None

        # Assume common defaults here before recalculating in set_formatting().
        self._default_time_width: int = 17
        self.column_widths['time'] = self._default_time_width
        self.column_widths['level'] = 3
        self._year_month_day_width: int = 9

        # Width of all columns except the final message
        self.column_width_prefix_total = 0

    def set_prefs(self, prefs: ConsolePrefs) -> None:
        self.prefs = prefs
        # Max column widths of each log field
        self.column_padding = ' ' * self.prefs.spaces_between_columns

    def all_column_names(self):
        columns_names = [name for name, _width in self._ordered_column_widths()]
        return columns_names + ['message']

    def _width_of_justified_fields(self):
        """Calculate the width of all columns except LAST_TABLE_COLUMN_NAMES."""
        padding_width = len(self.column_padding)
        used_width = sum(
            [
                width + padding_width
                for key, width in self.column_widths.items()
                if key not in TableView.LAST_TABLE_COLUMN_NAMES
            ]
        )
        return used_width

    def _ordered_column_widths(self):
        """Return each column and width in the preferred order."""
        if self.prefs.column_order:
            # Get ordered_columns
            columns = copy.copy(self.column_widths)
            ordered_columns = {}

            for column_name in self.prefs.column_order:
                # If valid column name
                if column_name in columns:
                    ordered_columns[column_name] = columns.pop(column_name)

            # Add remaining columns unless user preference to hide them.
            if not self.prefs.omit_unspecified_columns:
                for column_name in columns:
                    ordered_columns[column_name] = columns[column_name]
        else:
            ordered_columns = copy.copy(self.column_widths)

        if not self.prefs.show_python_file and 'py_file' in ordered_columns:
            del ordered_columns['py_file']
        if not self.prefs.show_python_logger and 'py_logger' in ordered_columns:
            del ordered_columns['py_logger']
        if not self.prefs.show_source_file and 'file' in ordered_columns:
            del ordered_columns['file']

        return ordered_columns.items()

    def update_metadata_column_widths(self, log: LogLine) -> None:
        """Calculate the max widths for each metadata field."""
        if log.metadata is None:
            log.update_metadata()
        # If extra fields still don't exist, no need to update column widths.
        if log.metadata is None:
            return

        for field_name, value in log.metadata.fields.items():
            value_string = str(value)

            # Get width of formatted numbers
            if isinstance(value, float):
                value_string = TableView.FLOAT_FORMAT % value
            elif isinstance(value, int):
                value_string = TableView.INT_FORMAT % value

            current_width = self.column_widths.get(field_name, 0)
            if len(value_string) > current_width:
                self.column_widths[field_name] = len(value_string)

        # Update log level character width.
        ansi_stripped_level = strip_ansi(log.record.levelname)
        if len(ansi_stripped_level) > self.column_widths['level']:
            self.column_widths['level'] = len(ansi_stripped_level)

        self.column_width_prefix_total = self._width_of_justified_fields()
        self._update_table_header()

    def _update_table_header(self):
        default_style = 'bold'
        fragments: collections.deque = collections.deque()

        # Update time column width to current prefs setting
        self.column_widths['time'] = self._default_time_width
        if self.prefs.hide_date_from_log_time:
            self.column_widths['time'] = (
                self._default_time_width - self._year_month_day_width
            )

        for name, width in self._ordered_column_widths():
            # These fields will be shown at the end
            if name in TableView.LAST_TABLE_COLUMN_NAMES:
                continue
            fragments.append((default_style, name.title()[:width].ljust(width)))
            fragments.append(('', self.column_padding))

        fragments.append((default_style, 'Message'))

        self._header_fragment_cache = list(fragments)

    def formatted_header(self):
        """Get pre-formatted table header."""
        return self._header_fragment_cache

    def formatted_row(self, log: LogLine) -> StyleAndTextTuples:
        """Render a single table row."""
        # pylint: disable=too-many-locals
        padding_formatted_text = ('', self.column_padding)
        # Don't apply any background styling that would override the parent
        # window or selected-log-line style.
        default_style = ''

        table_fragments: StyleAndTextTuples = []

        # NOTE: To preseve ANSI formatting on log level use:
        # table_fragments.extend(
        #     ANSI(log.record.levelname.ljust(
        #         self.column_widths['level'])).__pt_formatted_text__())

        # Collect remaining columns to display after host time and level.
        columns = {}
        for name, width in self._ordered_column_widths():
            # Skip these modifying these fields
            if name in TableView.LAST_TABLE_COLUMN_NAMES:
                continue

            # hasattr checks are performed here since a log record may not have
            # asctime or levelname if they are not included in the formatter
            # fmt string.
            if name == 'time' and hasattr(log.record, 'asctime'):
                time_text = log.record.asctime
                if self.prefs.hide_date_from_log_time:
                    time_text = time_text[self._year_month_day_width :]
                time_style = self.prefs.column_style(
                    'time', time_text, default='class:log-time'
                )
                columns['time'] = (
                    time_style,
                    time_text.ljust(self.column_widths['time']),
                )
                continue

            if name == 'level' and hasattr(log.record, 'levelname'):
                # Remove any existing ANSI formatting and apply our colors.
                level_text = strip_ansi(log.record.levelname)
                level_style = self.prefs.column_style(
                    'level',
                    level_text,
                    default='class:log-level-{}'.format(log.record.levelno),
                )
                columns['level'] = (
                    level_style,
                    level_text.ljust(self.column_widths['level']),
                )
                continue

            value = log.metadata.fields.get(name, ' ')
            left_justify = True

            # Right justify and format numbers
            if isinstance(value, float):
                value = TableView.FLOAT_FORMAT % value
                left_justify = False
            elif isinstance(value, int):
                value = TableView.INT_FORMAT % value
                left_justify = False

            if left_justify:
                columns[name] = value.ljust(width)
            else:
                columns[name] = value.rjust(width)

        # Grab the message to appear after the justified columns with ANSI
        # escape sequences removed.
        message_text = strip_ansi(log.record.message)
        message = log.metadata.fields.get(
            'msg',
            message_text.rstrip(),  # Remove any trailing line breaks
        )
        # Alternatively ANSI formatting can be preserved with:
        #   message = ANSI(log.record.message).__pt_formatted_text__()

        # Convert to FormattedText if we have a raw string from fields.
        if isinstance(message, str):
            message_style = default_style
            if log.record.levelno >= 30:  # Warning, Error and Critical
                # Style the whole message to match it's level
                message_style = 'class:log-level-{}'.format(log.record.levelno)
            message = (message_style, message)
        # Add to columns
        columns['message'] = message

        index_modifier = 0
        # Go through columns and convert to FormattedText where needed.
        for i, column in enumerate(columns.items()):
            column_name, column_value = column
            if i in [0, 1] and column_name in ['time', 'level']:
                index_modifier -= 1
            # For raw strings that don't have their own ANSI colors, apply the
            # theme color style for this column.
            if isinstance(column_value, str):
                fallback_style = (
                    'class:log-table-column-{}'.format(i + index_modifier)
                    if 0 <= i <= 7
                    else default_style
                )

                style = self.prefs.column_style(
                    column_name, column_value.rstrip(), default=fallback_style
                )

                table_fragments.append((style, column_value))
                table_fragments.append(padding_formatted_text)
            # Add this tuple to table_fragments.
            elif isinstance(column, tuple):
                table_fragments.append(column_value)
                # Add padding if not the last column.
                if i < len(columns) - 1:
                    table_fragments.append(padding_formatted_text)

        # Add the final new line for this row.
        table_fragments.append(('', '\n'))
        return table_fragments



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Script that invokes protoc to generate code for .proto files."""

import argparse
import logging
import os
from pathlib import Path
import subprocess
import sys
import tempfile
from typing import Callable, Dict, List, Optional, Tuple, Union

# Make sure dependencies are optional, since this script may be run when
# installing Python package dependencies through GN.
try:
    from pw_cli.log import install as setup_logging
except ImportError:
    from logging import basicConfig as setup_logging  # type: ignore

_LOG = logging.getLogger(__name__)


def _argument_parser() -> argparse.ArgumentParser:
    """Registers the script's arguments on an argument parser."""

    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--language',
        required=True,
        choices=DEFAULT_PROTOC_ARGS,
        help='Output language',
    )
    parser.add_argument(
        '--plugin-path', type=Path, help='Path to the protoc plugin'
    )
    parser.add_argument(
        '--proto-path',
        type=Path,
        help='Additional protoc include paths',
        action='append',
    )
    parser.add_argument(
        '--include-file',
        type=argparse.FileType('r'),
        help='File containing additional protoc include paths',
    )
    parser.add_argument(
        '--out-dir',
        type=Path,
        required=True,
        help='Output directory for generated code',
    )
    parser.add_argument(
        '--compile-dir',
        type=Path,
        required=True,
        help='Root path for compilation',
    )
    parser.add_argument(
        '--sources', type=Path, nargs='+', help='Input protobuf files'
    )
    parser.add_argument(
        '--protoc', type=Path, default='protoc', help='Path to protoc'
    )
    parser.add_argument(
        '--no-experimental-proto3-optional',
        dest='experimental_proto3_optional',
        action='store_false',
        help='Do not invoke protoc with --experimental_allow_proto3_optional',
    )
    parser.add_argument(
        '--no-generate-type-hints',
        dest='generate_type_hints',
        action='store_false',
        help='Do not generate pyi files for python',
    )
    parser.add_argument(
        '--exclude-pwpb-legacy-snake-case-field-name-enums',
        dest='exclude_pwpb_legacy_snake_case_field_name_enums',
        action='store_true',
        help=(
            'If set, generates legacy SNAKE_CASE names for field name enums '
            'in PWPB.'
        ),
    )

    return parser


def protoc_common_args(args: argparse.Namespace) -> Tuple[str, ...]:
    flags: Tuple[str, ...] = ()
    if args.experimental_proto3_optional:
        flags += ('--experimental_allow_proto3_optional',)
    return flags


def protoc_pwpb_args(
    args: argparse.Namespace, include_paths: List[str]
) -> Tuple[str, ...]:
    out_args = [
        '--plugin',
        f'protoc-gen-custom={args.plugin_path}',
        f'--custom_opt=-I{args.compile_dir}',
        *[f'--custom_opt=-I{include_path}' for include_path in include_paths],
    ]

    if args.exclude_pwpb_legacy_snake_case_field_name_enums:
        out_args.append(
            '--custom_opt=--exclude-legacy-snake-case-field-name-enums'
        )

    out_args.extend(
        [
            '--custom_out',
            args.out_dir,
        ]
    )

    return tuple(out_args)


def protoc_pwpb_rpc_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    return (
        '--plugin',
        f'protoc-gen-custom={args.plugin_path}',
        '--custom_out',
        args.out_dir,
    )


def protoc_go_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    return (
        '--go_out',
        f'plugins=grpc:{args.out_dir}',
    )


def protoc_nanopb_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    # nanopb needs to know of the include path to parse *.options files
    return (
        '--plugin',
        f'protoc-gen-nanopb={args.plugin_path}',
        # nanopb_opt provides the flags to use for nanopb_out. Windows doesn't
        # like when you merge the two using the `flag,...:out` syntax. Use
        # Posix-style paths since backslashes on Windows are treated like
        # escape characters.
        f'--nanopb_opt=-I{args.compile_dir.as_posix()}',
        f'--nanopb_out={args.out_dir}',
    )


def protoc_nanopb_rpc_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    return (
        '--plugin',
        f'protoc-gen-custom={args.plugin_path}',
        '--custom_out',
        args.out_dir,
    )


def protoc_raw_rpc_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    return (
        '--plugin',
        f'protoc-gen-custom={args.plugin_path}',
        '--custom_out',
        args.out_dir,
    )


def protoc_python_args(
    args: argparse.Namespace, _include_paths: List[str]
) -> Tuple[str, ...]:
    flags: Tuple[str, ...] = (
        '--python_out',
        args.out_dir,
    )

    if args.generate_type_hints:
        flags += (
            '--mypy_out',
            args.out_dir,
        )

    return flags


_DefaultArgsFunction = Callable[
    [argparse.Namespace, List[str]], Tuple[str, ...]
]

# Default additional protoc arguments for each supported language.
# TODO(frolv): Make these overridable with a command-line argument.
DEFAULT_PROTOC_ARGS: Dict[str, _DefaultArgsFunction] = {
    'go': protoc_go_args,
    'nanopb': protoc_nanopb_args,
    'nanopb_rpc': protoc_nanopb_rpc_args,
    'pwpb': protoc_pwpb_args,
    'pwpb_rpc': protoc_pwpb_rpc_args,
    'python': protoc_python_args,
    'raw_rpc': protoc_raw_rpc_args,
}

# Languages that protoc internally supports.
BUILTIN_PROTOC_LANGS = ('go', 'python')


def main() -> int:
    """Runs protoc as configured by command-line arguments."""

    parser = _argument_parser()
    args = parser.parse_args()

    if args.plugin_path is None and args.language not in BUILTIN_PROTOC_LANGS:
        parser.error(
            f'--plugin-path is required for --language {args.language}'
        )

    args.out_dir.mkdir(parents=True, exist_ok=True)

    include_paths: List[str] = []
    if args.include_file:
        include_paths.extend(line.strip() for line in args.include_file)
    if args.proto_path:
        include_paths.extend(str(path) for path in args.proto_path)

    wrapper_script: Optional[Path] = None

    # On Windows, use a .bat version of the plugin if it exists or create a .bat
    # wrapper to use if none exists.
    if os.name == 'nt' and args.plugin_path:
        if args.plugin_path.with_suffix('.bat').exists():
            args.plugin_path = args.plugin_path.with_suffix('.bat')
            _LOG.debug('Using Batch plugin %s', args.plugin_path)
        else:
            with tempfile.NamedTemporaryFile(
                'w', suffix='.bat', delete=False
            ) as file:
                file.write(f'@echo off\npython {args.plugin_path.resolve()}\n')

            args.plugin_path = wrapper_script = Path(file.name)
            _LOG.debug('Using generated plugin wrapper %s', args.plugin_path)

    cmd: Tuple[Union[str, Path], ...] = (
        args.protoc,
        f'-I{args.compile_dir}',
        *[f'-I{include_path}' for include_path in include_paths],
        *protoc_common_args(args),
        *DEFAULT_PROTOC_ARGS[args.language](args, include_paths),
        *args.sources,
    )

    try:
        process = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    finally:
        if wrapper_script:
            wrapper_script.unlink()

    if process.returncode != 0:
        _LOG.error(
            'Protocol buffer compilation failed!\n%s',
            ' '.join(str(c) for c in cmd),
        )
        sys.stderr.buffer.write(process.stdout)
        sys.stderr.flush()

    return process.returncode


if __name__ == '__main__':
    setup_logging()
    sys.exit(main())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Functions to create checkboxes for menus and toolbars."""

import sys
from typing import Callable, Iterable, Optional, NamedTuple

from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
from prompt_toolkit.formatted_text import StyleAndTextTuples

_KEY_SEPARATOR = ' '
_CHECKED_BOX = '[✓]'

if sys.platform in ['win32']:
    _CHECKED_BOX = '[x]'


class ToolbarButton(NamedTuple):
    key: Optional[str] = None
    description: Optional[str] = 'Button'
    mouse_handler: Optional[Callable] = None
    is_checkbox: bool = False
    checked: Optional[Callable] = None


def to_checkbox(
    checked: bool,
    mouse_handler: Optional[Callable] = None,
    end: str = ' ',
    unchecked_style: str = 'class:checkbox',
    checked_style: str = 'class:checkbox-checked',
) -> OneStyleAndTextTuple:
    text = _CHECKED_BOX if checked else '[ ]'
    text += end
    style = checked_style if checked else unchecked_style
    if mouse_handler:
        return (style, text, mouse_handler)
    return (style, text)


def to_checkbox_text(checked: bool, end=' '):
    return to_checkbox(checked, end=end)[1]


def to_setting(
    checked: bool,
    text: str,
    active_style='class:toolbar-setting-active',
    inactive_style='',
    mouse_handler=None,
):
    """Apply a style to text if checked is True."""
    style = active_style if checked else inactive_style
    if mouse_handler:
        return (style, text, mouse_handler)
    return (style, text)


def to_checkbox_with_keybind_indicator(
    checked: bool,
    key: str,
    description: str,
    mouse_handler=None,
    base_style: str = '',
    **checkbox_kwargs,
):
    """Create a clickable keybind indicator with checkbox for toolbars."""
    if mouse_handler:
        return to_keybind_indicator(
            key,
            description,
            mouse_handler,
            leading_fragments=[
                to_checkbox(checked, mouse_handler, **checkbox_kwargs)
            ],
            base_style=base_style,
        )
    return to_keybind_indicator(
        key,
        description,
        leading_fragments=[to_checkbox(checked, **checkbox_kwargs)],
        base_style=base_style,
    )


def to_keybind_indicator(
    key: str,
    description: str,
    mouse_handler: Optional[Callable] = None,
    leading_fragments: Optional[Iterable] = None,
    middle_fragments: Optional[Iterable] = None,
    base_style: str = '',
    key_style: str = 'class:keybind',
    description_style: str = 'class:keyhelp',
):
    """Create a clickable keybind indicator for toolbars."""
    if base_style:
        base_style += ' '

    fragments: StyleAndTextTuples = []
    fragments.append((base_style + 'class:toolbar-button-decoration', ' '))

    def append_fragment_with_base_style(frag_list, fragment) -> None:
        if mouse_handler:
            frag_list.append(
                (base_style + fragment[0], fragment[1], mouse_handler)
            )
        else:
            frag_list.append((base_style + fragment[0], fragment[1]))

    # Add any starting fragments first
    if leading_fragments:
        for fragment in leading_fragments:
            append_fragment_with_base_style(fragments, fragment)

    # Function name
    if mouse_handler:
        fragments.append(
            (base_style + description_style, description, mouse_handler)
        )
    else:
        fragments.append((base_style + description_style, description))

    if middle_fragments:
        for fragment in middle_fragments:
            append_fragment_with_base_style(fragments, fragment)

    # Separator and keybind
    if key:
        if mouse_handler:
            fragments.append(
                (base_style + description_style, _KEY_SEPARATOR, mouse_handler)
            )
            fragments.append((base_style + key_style, key, mouse_handler))
        else:
            fragments.append((base_style + description_style, _KEY_SEPARATOR))
            fragments.append((base_style + key_style, key))

    fragments.append((base_style + 'class:toolbar-button-decoration', ' '))
    return fragments



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Module for size report ASCII tables from DataSourceMaps."""

import enum
from typing import (
    Iterable,
    Tuple,
    Union,
    Type,
    List,
    Optional,
    NamedTuple,
    cast,
)

from pw_bloat.label import DataSourceMap, DiffDataSourceMap, Label


class AsciiCharset(enum.Enum):
    """Set of ASCII characters for drawing tables."""

    TL = '+'
    TM = '+'
    TR = '+'
    ML = '+'
    MM = '+'
    MR = '+'
    BL = '+'
    BM = '+'
    BR = '+'
    V = '|'
    H = '-'
    HH = '='


class LineCharset(enum.Enum):
    """Set of line-drawing characters for tables."""

    TL = '┌'
    TM = '┬'
    TR = '┐'
    ML = '├'
    MM = '┼'
    MR = '┤'
    BL = '└'
    BM = '┴'
    BR = '┘'
    V = '│'
    H = '─'
    HH = '═'


class _Align(enum.Enum):
    CENTER = 0
    LEFT = 1
    RIGHT = 2


def get_label_status(curr_label: Label) -> str:
    if curr_label.is_new():
        return 'NEW'
    if curr_label.is_del():
        return 'DEL'
    return ''


def diff_sign_sizes(size: int, diff_mode: bool) -> str:
    if diff_mode:
        size_sign = '+' if size > 0 else ''
        return f"{size_sign}{size:,}"
    return f"{size:,}"


class BloatTableOutput:
    """ASCII Table generator from DataSourceMap."""

    _RST_PADDING_WIDTH = 6
    _DEFAULT_MAX_WIDTH = 80

    class _LabelContent(NamedTuple):
        name: str
        size: int
        label_status: str

    def __init__(
        self,
        ds_map: Union[DiffDataSourceMap, DataSourceMap],
        col_max_width: int = _DEFAULT_MAX_WIDTH,
        charset: Union[Type[AsciiCharset], Type[LineCharset]] = AsciiCharset,
        rst_output: bool = False,
        diff_label: Optional[str] = None,
    ):
        self._data_source_map = ds_map
        self._cs = charset
        self._total_size = 0
        col_names = [*self._data_source_map.get_ds_names(), 'sizes']
        self._diff_mode = False
        self._diff_label = diff_label
        if isinstance(self._data_source_map, DiffDataSourceMap):
            col_names = ['diff', *col_names]
            self._diff_mode = True
        self._col_names = col_names
        self._additional_padding = 0
        self._ascii_table_rows: List[str] = []
        self._rst_output = rst_output
        self._total_divider = self._cs.HH.value
        if self._rst_output:
            self._total_divider = self._cs.H.value
            self._additional_padding = self._RST_PADDING_WIDTH

        self._col_widths = self._generate_col_width(col_max_width)

    def _generate_col_width(self, col_max_width: int) -> List[int]:
        """Find column width for all data sources and sizes."""
        max_len_size = 0
        diff_len_col_width = 0

        col_list = [
            len(ds_name) for ds_name in self._data_source_map.get_ds_names()
        ]
        for curr_label in self._data_source_map.labels():
            self._total_size += curr_label.size
            max_len_size = max(
                len(diff_sign_sizes(self._total_size, self._diff_mode)),
                len(diff_sign_sizes(curr_label.size, self._diff_mode)),
                max_len_size,
            )
            for index, parent_label in enumerate(
                [*curr_label.parents, curr_label.name]
            ):
                if len(parent_label) > col_max_width:
                    col_list[index] = col_max_width
                elif len(parent_label) > col_list[index]:
                    col_list[index] = len(parent_label)

        diff_same = 0
        if self._diff_mode:
            col_list = [len('Total'), *col_list]
            diff_same = len('(SAME)')
        col_list.append(max(max_len_size, len('sizes'), diff_same))

        if self._diff_label is not None:
            sum_all_col_names = sum(col_list)
            if sum_all_col_names < len(self._diff_label):
                diff_len_col_width = (
                    len(self._diff_label) - sum_all_col_names
                ) // len(self._col_names)

        return [
            (x + self._additional_padding + diff_len_col_width)
            for x in col_list
        ]

    def _diff_label_names(
        self,
        old_labels: Optional[Tuple[_LabelContent, ...]],
        new_labels: Tuple[_LabelContent, ...],
    ) -> Tuple[_LabelContent, ...]:
        """Return difference between arrays of labels."""

        if old_labels is None:
            return new_labels
        diff_list = []
        for new_lb, old_lb in zip(new_labels, old_labels):
            if (new_lb.name == old_lb.name) and (new_lb.size == old_lb.size):
                diff_list.append(self._LabelContent('', 0, ''))
            else:
                diff_list.append(new_lb)

        return tuple(diff_list)

    def _label_title_row(self) -> List[str]:
        label_rows = []
        label_cells = ''
        divider_cells = ''
        for width in self._col_widths:
            label_cells += ' ' * width + ' '
            divider_cells += (self._cs.H.value * width) + self._cs.H.value
        if self._diff_label is not None:
            label_cells = self._diff_label.center(len(label_cells[:-1]), ' ')
        label_rows.extend(
            [
                f"{self._cs.TL.value}{divider_cells[:-1]}{self._cs.TR.value}",
                f"{self._cs.V.value}{label_cells}{self._cs.V.value}",
                f"{self._cs.ML.value}{divider_cells[:-1]}{self._cs.MR.value}",
            ]
        )
        return label_rows

    def create_table(self) -> str:
        """Parse DataSourceMap to create ASCII table."""
        curr_lb_hierachy = None
        last_diff_name = ''
        if self._diff_mode:
            self._ascii_table_rows.extend([*self._label_title_row()])
        else:
            self._ascii_table_rows.extend(
                [self._create_border(True, self._cs.H.value)]
            )
        self._ascii_table_rows.extend([*self._create_title_row()])

        has_entries = False

        for curr_label in self._data_source_map.labels():
            if curr_label.size == 0:
                continue

            has_entries = True

            new_lb_hierachy = tuple(
                [
                    *self._get_ds_label_size(curr_label.parents),
                    self._LabelContent(
                        curr_label.name,
                        curr_label.size,
                        get_label_status(curr_label),
                    ),
                ]
            )
            diff_list = self._diff_label_names(
                curr_lb_hierachy, new_lb_hierachy
            )
            curr_lb_hierachy = new_lb_hierachy

            if curr_label.parents and curr_label.parents[0] == last_diff_name:
                continue
            if (
                self._diff_mode
                and diff_list[0].name
                and (
                    not cast(
                        DiffDataSourceMap, self._data_source_map
                    ).has_diff_sublabels(diff_list[0].name)
                )
            ):
                if (len(self._ascii_table_rows) > 5) and (
                    self._ascii_table_rows[-1][0] != '+'
                ):
                    self._ascii_table_rows.append(
                        self._row_divider(
                            len(self._col_names), self._cs.H.value
                        )
                    )
                self._ascii_table_rows.append(
                    self._create_same_label_row(1, diff_list[0].name)
                )

                last_diff_name = curr_label.parents[0]
            else:
                self._ascii_table_rows += self._create_diff_rows(diff_list)

        if self._rst_output and self._ascii_table_rows[-1][0] == '+':
            self._ascii_table_rows.pop()

        self._ascii_table_rows.extend(
            [*self._create_total_row(is_empty=not has_entries)]
        )

        return '\n'.join(self._ascii_table_rows) + '\n'

    def _create_same_label_row(self, col_index: int, label: str) -> str:
        label_row = ''
        for col in range(len(self._col_names) - 1):
            if col == col_index:
                curr_cell = self._create_cell(label, False, col, _Align.LEFT)
            else:
                curr_cell = self._create_cell('', False, col)
            label_row += curr_cell
        label_row += self._create_cell(
            "(SAME)", True, len(self._col_widths) - 1, _Align.RIGHT
        )
        return label_row

    def _get_ds_label_size(
        self, parent_labels: Tuple[str, ...]
    ) -> Iterable[_LabelContent]:
        """Produce label, size pairs from parent label names."""
        parent_label_sizes = []
        for index, target_label in enumerate(parent_labels):
            for curr_label in self._data_source_map.labels(index):
                if curr_label.name == target_label:
                    diff_label = get_label_status(curr_label)
                    parent_label_sizes.append(
                        self._LabelContent(
                            curr_label.name, curr_label.size, diff_label
                        )
                    )
                    break
        return parent_label_sizes

    def _create_total_row(self, is_empty: bool = False) -> Iterable[str]:
        complete_total_rows = []

        if self._diff_mode and is_empty:
            # When diffing two identical binaries, output a row indicating that
            # the two are the same.
            no_diff_row = ''
            for i in range(len(self._col_names)):
                if i == 0:
                    no_diff_row += self._create_cell(
                        'N/A', False, i, _Align.CENTER
                    )
                elif i == len(self._col_names) - 1:
                    no_diff_row += self._create_cell('0', True, i)
                else:
                    no_diff_row += self._create_cell(
                        '(same)', False, i, _Align.CENTER
                    )
            complete_total_rows.append(no_diff_row)

        complete_total_rows.append(
            self._row_divider(len(self._col_names), self._total_divider)
        )
        total_row = ''

        for i in range(len(self._col_names)):
            if i == 0:
                total_row += self._create_cell('Total', False, i, _Align.LEFT)
            elif i == len(self._col_names) - 1:
                total_size_str = diff_sign_sizes(
                    self._total_size, self._diff_mode
                )
                total_row += self._create_cell(total_size_str, True, i)
            else:
                total_row += self._create_cell('', False, i, _Align.CENTER)

        complete_total_rows.extend(
            [total_row, self._create_border(False, self._cs.H.value)]
        )
        return complete_total_rows

    def _create_diff_rows(
        self, diff_list: Tuple[_LabelContent, ...]
    ) -> Iterable[str]:
        """Create rows for each label according to its index in diff_list."""
        curr_row = ''
        diff_index = 0
        diff_rows = []
        for index, label_content in enumerate(diff_list):
            if label_content.name:
                if self._diff_mode:
                    curr_row += self._create_cell(
                        label_content.label_status, False, 0
                    )
                    diff_index = 1
                for cell_index in range(
                    diff_index, len(diff_list) + diff_index
                ):
                    if cell_index == index + diff_index:
                        if (
                            cell_index == diff_index
                            and len(self._ascii_table_rows) > 5
                            and not self._rst_output
                        ):
                            diff_rows.append(
                                self._row_divider(
                                    len(self._col_names), self._cs.H.value
                                )
                            )
                        if (
                            len(label_content.name) + self._additional_padding
                        ) > self._col_widths[cell_index]:
                            curr_row = self._multi_row_label(
                                label_content.name, cell_index
                            )
                            break
                        curr_row += self._create_cell(
                            label_content.name, False, cell_index, _Align.LEFT
                        )
                    else:
                        curr_row += self._create_cell('', False, cell_index)

                # Add size end of current row.
                curr_size = diff_sign_sizes(label_content.size, self._diff_mode)
                curr_row += self._create_cell(
                    curr_size, True, len(self._col_widths) - 1, _Align.RIGHT
                )
                diff_rows.append(curr_row)
                if self._rst_output:
                    diff_rows.append(
                        self._row_divider(
                            len(self._col_names), self._cs.H.value
                        )
                    )
                curr_row = ''

        return diff_rows

    def _create_cell(
        self,
        content: str,
        last_cell: bool,
        col_index: int,
        align: Optional[_Align] = _Align.RIGHT,
    ) -> str:
        v_border = self._cs.V.value
        if self._rst_output and content:
            content = content.replace('_', '\\_')
        pad_diff = self._col_widths[col_index] - len(content)
        padding = (pad_diff // 2) * ' '
        odd_pad = ' ' if pad_diff % 2 == 1 else ''
        string_cell = ''

        if align == _Align.CENTER:
            string_cell = f'{v_border}{odd_pad}{padding}{content}{padding}'
        elif align == _Align.LEFT:
            string_cell = f'{v_border}{content}{padding*2}{odd_pad}'
        elif align == _Align.RIGHT:
            string_cell = f'{v_border}{padding*2}{odd_pad}{content}'

        if last_cell:
            string_cell += self._cs.V.value
        return string_cell

    def _multi_row_label(self, content: str, target_col_index: int) -> str:
        """Split content name into multiple rows within correct column."""
        max_len = self._col_widths[target_col_index] - self._additional_padding
        split_content = '...'.join(
            content[max_len:][i : i + max_len - 3]
            for i in range(0, len(content[max_len:]), max_len - 3)
        )
        split_content = f"{content[:max_len]}...{split_content}"
        split_tab_content = [
            split_content[i : i + max_len]
            for i in range(0, len(split_content), max_len)
        ]
        multi_label = []
        curr_row = ''
        for index, cut_content in enumerate(split_tab_content):
            last_cell = False
            for blank_cell_index in range(len(self._col_names)):
                if blank_cell_index == target_col_index:
                    curr_row += self._create_cell(
                        cut_content, False, target_col_index, _Align.LEFT
                    )
                else:
                    if blank_cell_index == len(self._col_names) - 1:
                        if index == len(split_tab_content) - 1:
                            break
                        last_cell = True
                    curr_row += self._create_cell(
                        '', last_cell, blank_cell_index
                    )
            multi_label.append(curr_row)
            curr_row = ''

        return '\n'.join(multi_label)

    def _row_divider(self, col_num: int, h_div: str) -> str:
        l_border = ''
        r_border = ''
        row_div = ''
        for col in range(col_num):
            if col == 0:
                l_border = self._cs.ML.value
                r_border = ''
            elif col == (col_num - 1):
                l_border = self._cs.MM.value
                r_border = self._cs.MR.value
            else:
                l_border = self._cs.MM.value
                r_border = ''

            row_div += f"{l_border}{self._col_widths[col] * h_div}{r_border}"
        return row_div

    def _create_title_row(self) -> Iterable[str]:
        title_rows = []
        title_cells = ''
        last_cell = False
        for index, curr_name in enumerate(self._col_names):
            if index == len(self._col_names) - 1:
                last_cell = True
            title_cells += self._create_cell(
                curr_name, last_cell, index, _Align.CENTER
            )
        title_rows.extend(
            [
                title_cells,
                self._row_divider(len(self._col_names), self._cs.HH.value),
            ]
        )
        return title_rows

    def _create_border(self, top: bool, h_div: str):
        """Top or bottom borders of ASCII table."""
        row_div = ''
        for col in range(len(self._col_names)):
            if top:
                if col == 0:
                    l_div = self._cs.TL.value
                    r_div = ''
                elif col == (len(self._col_names) - 1):
                    l_div = self._cs.TM.value
                    r_div = self._cs.TR.value
                else:
                    l_div = self._cs.TM.value
                    r_div = ''
            else:
                if col == 0:
                    l_div = self._cs.BL.value
                    r_div = ''
                elif col == (len(self._col_names) - 1):
                    l_div = self._cs.BM.value
                    r_div = self._cs.BR.value
                else:
                    l_div = self._cs.BM.value
                    r_div = ''

            row_div += f"{l_div}{self._col_widths[col] * h_div}{r_div}"
        return row_div


class RstOutput:
    """Tabular output in ASCII format, which is also valid RST."""

    def __init__(
        self, ds_map: DataSourceMap, table_label: Optional[str] = None
    ):
        self._data_source_map = ds_map
        self._table_label = table_label
        self._diff_mode = False
        if isinstance(self._data_source_map, DiffDataSourceMap):
            self._diff_mode = True

    def create_table(self) -> str:
        """Initializes RST table and builds first row."""
        table_builder = [
            '\n.. list-table::',
            '   :widths: auto',
            '   :header-rows: 1\n',
        ]
        header_cols = ['Label', 'Segment', 'Delta']
        for i, col_name in enumerate(header_cols):
            list_space = '*' if i == 0 else ' '
            table_builder.append(f"   {list_space} - {col_name}")

        return '\n'.join(table_builder) + f'\n{self.add_report_row()}\n'

    def _label_status_unchanged(self, parent_lb_name: str) -> bool:
        """Determines if parent label has no status change in diff mode."""
        for curr_lb in self._data_source_map.labels():
            if curr_lb.size != 0:
                if (
                    curr_lb.parents and (parent_lb_name == curr_lb.parents[0])
                ) or (curr_lb.name == parent_lb_name):
                    if get_label_status(curr_lb) != '':
                        return False
        return True

    def add_report_row(self) -> str:
        """Add in new size report row with Label, Segment, and Delta.

        Returns:
            RST string that is the current row with a full symbols
            table breakdown of the corresponding segment.
        """
        table_rows = []
        curr_row = []
        curr_label_name = ''
        for parent_lb in self._data_source_map.labels(0):
            if parent_lb.size != 0:
                if (self._table_label is not None) and (
                    curr_label_name != self._table_label
                ):
                    curr_row.append(f'   * - {self._table_label} ')
                    curr_label_name = self._table_label
                else:
                    curr_row.append('   * -')
                curr_row.extend(
                    [
                        f'     - .. dropdown:: {parent_lb.name}',
                        '            :animate: fade-in\n',
                        '            .. list-table::',
                        '               :widths: auto\n',
                    ]
                )
                if self._label_status_unchanged(parent_lb.name):
                    skip_status = 1, '*'
                else:
                    skip_status = 0, ' '
                for curr_lb in self._data_source_map.labels():
                    if (curr_lb.size != 0) and (
                        (
                            curr_lb.parents
                            and (parent_lb.name == curr_lb.parents[0])
                        )
                        or (curr_lb.name == parent_lb.name)
                    ):
                        sign_size = diff_sign_sizes(
                            curr_lb.size, self._diff_mode
                        )
                        curr_status = get_label_status(curr_lb)
                        curr_name = curr_lb.name.replace('_', '\\_')
                        to_extend = [
                            f'               * - {curr_status}',
                            f'               {skip_status[1]} - {sign_size}',
                            f'                 - {curr_name}\n',
                        ][skip_status[0] :]
                        curr_row.extend(to_extend)
                curr_row.append(
                    f'     - {diff_sign_sizes(parent_lb.size, self._diff_mode)}'
                )
            table_rows.extend(curr_row)
            curr_row = []

        # No size difference.
        if len(table_rows) == 0:
            table_rows.extend(
                [f'\n   * - {self._table_label}', '     - (ALL)', '     - 0']
            )

        return '\n'.join(table_rows)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Code for improving interactive use of Python functions."""

import inspect
import textwrap
from typing import Callable


def _annotation_name(annotation: object) -> str:
    if isinstance(annotation, str):
        return annotation

    return getattr(annotation, '__name__', repr(annotation))


def format_parameter(param: inspect.Parameter) -> str:
    """Formats a parameter for printing in a function signature."""
    if param.kind == param.VAR_POSITIONAL:
        name = '*' + param.name
    elif param.kind == param.VAR_KEYWORD:
        name = '**' + param.name
    else:
        name = param.name

    if param.default is param.empty:
        default = ''
    else:
        default = f' = {param.default}'

    if param.annotation is param.empty:
        annotation = ''
    else:
        annotation = f': {_annotation_name(param.annotation)}'

    return f'{name}{annotation}{default}'


def format_signature(name: str, signature: inspect.Signature) -> str:
    """Formats a function signature as if it were source code.

    Does not yet handle / and * markers.
    """
    params = ', '.join(
        format_parameter(arg) for arg in signature.parameters.values()
    )
    if signature.return_annotation is signature.empty:
        return_annotation = ''
    else:
        return_annotation = ' -> ' + _annotation_name(
            signature.return_annotation
        )

    return f'{name}({params}){return_annotation}'


def format_function_help(function: Callable) -> str:
    """Formats a help string with a declaration and docstring."""
    signature = format_signature(
        function.__name__, inspect.signature(function, follow_wrapped=False)
    )

    docs = inspect.getdoc(function) or '(no docstring)'
    return f'{signature}:\n\n{textwrap.indent(docs, "    ")}'


def help_as_repr(function: Callable) -> Callable:
    """Wraps a function so that its repr() and docstring provide detailed help.

    This is useful for creating commands in an interactive console. In a
    console, typing a function's name and hitting Enter shows rich documentation
    with the full function signature, type annotations, and docstring when the
    function is wrapped with help_as_repr.
    """

    def display_help(_):
        return format_function_help(function)

    return type(
        function.__name__,
        (),
        dict(
            __call__=staticmethod(function),
            __doc__=format_function_help(function),
            __repr__=display_help,
        ),
    )()



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Options file parsing for proto generation."""

from fnmatch import fnmatchcase
from pathlib import Path
import re
from typing import List, Tuple

from google.protobuf import text_format

from pw_protobuf_codegen_protos.codegen_options_pb2 import CodegenOptions
from pw_protobuf_protos.field_options_pb2 import FieldOptions

_MULTI_LINE_COMMENT_RE = re.compile(r'/\*.*?\*/', flags=re.MULTILINE)
_SINGLE_LINE_COMMENT_RE = re.compile(r'//.*?$', flags=re.MULTILINE)
_SHELL_STYLE_COMMENT_RE = re.compile(r'#.*?$', flags=re.MULTILINE)

# A list of (proto field path, CodegenOptions) tuples.
ParsedOptions = List[Tuple[str, CodegenOptions]]


def load_options_from(options: ParsedOptions, options_file_name: Path):
    """Loads a single .options file for the given .proto"""
    with open(options_file_name) as options_file:
        # Read the options file and strip all styles of comments before parsing.
        options_data = options_file.read()
        options_data = _MULTI_LINE_COMMENT_RE.sub('', options_data)
        options_data = _SINGLE_LINE_COMMENT_RE.sub('', options_data)
        options_data = _SHELL_STYLE_COMMENT_RE.sub('', options_data)

        for line in options_data.split('\n'):
            parts = line.strip().split(None, 1)
            if len(parts) < 2:
                continue

            # Parse as a name glob followed by a protobuf text format.
            try:
                opts = CodegenOptions()
                text_format.Merge(parts[1], opts)
                options.append((parts[0], opts))
            except:  # pylint: disable=bare-except
                continue


def load_options(
    include_paths: List[Path], proto_file_name: Path
) -> ParsedOptions:
    """Loads the .options for the given .proto."""
    options: ParsedOptions = []

    for include_path in include_paths:
        options_file_name = include_path / proto_file_name.with_suffix(
            '.options'
        )
        if options_file_name.exists():
            load_options_from(options, options_file_name)

    return options


def match_options(name: str, options: ParsedOptions) -> CodegenOptions:
    """Return the matching options for a name."""
    matched = CodegenOptions()
    for name_glob, mask_options in options:
        if fnmatchcase(name, name_glob):
            matched.MergeFrom(mask_options)

    return matched


def create_from_field_options(
    field_options: FieldOptions,
) -> CodegenOptions:
    """Create a CodegenOptions from a FieldOptions."""
    codegen_options = CodegenOptions()

    if field_options.HasField('max_count'):
        codegen_options.max_count = field_options.max_count

    if field_options.HasField('max_size'):
        codegen_options.max_size = field_options.max_size

    return codegen_options


def merge_field_and_codegen_options(
    field_options: CodegenOptions, codegen_options: CodegenOptions
) -> CodegenOptions:
    """Merge inline field_options and options file codegen_options."""
    # The field options specify protocol-level requirements. Therefore, any
    # codegen options should not violate those protocol-level requirements.
    if field_options.max_count > 0 and codegen_options.max_count > 0:
        assert field_options.max_count == codegen_options.max_count

    if field_options.max_size > 0 and codegen_options.max_size > 0:
        assert field_options.max_size == codegen_options.max_size

    merged_options = CodegenOptions()
    merged_options.CopyFrom(field_options)
    merged_options.MergeFrom(codegen_options)

    return merged_options



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Library to analyze and dump Thread protos and Thread snapshots into text."""

import binascii
from typing import Optional, List, Mapping
import pw_tokenizer
from pw_symbolizer import LlvmSymbolizer, Symbolizer
from pw_tokenizer import proto as proto_detokenizer
from pw_thread_protos import thread_pb2

THREAD_STATE_TO_STRING: Mapping[int, str] = {
    thread_pb2.ThreadState.Enum.UNKNOWN: 'UNKNOWN',
    thread_pb2.ThreadState.Enum.INTERRUPT_HANDLER: 'INTERRUPT_HANDLER',
    thread_pb2.ThreadState.Enum.RUNNING: 'RUNNING',
    thread_pb2.ThreadState.Enum.READY: 'READY',
    thread_pb2.ThreadState.Enum.SUSPENDED: 'SUSPENDED',
    thread_pb2.ThreadState.Enum.BLOCKED: 'BLOCKED',
    thread_pb2.ThreadState.Enum.INACTIVE: 'INACTIVE',
}


def process_snapshot(
    serialized_snapshot: bytes,
    tokenizer_db: Optional[pw_tokenizer.Detokenizer] = None,
    symbolizer: Optional[Symbolizer] = None,
) -> str:
    """Processes snapshot threads, producing a multi-line string."""
    captured_threads = thread_pb2.SnapshotThreadInfo()
    captured_threads.ParseFromString(serialized_snapshot)
    if symbolizer is None:
        symbolizer = LlvmSymbolizer()

    return str(
        ThreadSnapshotAnalyzer(captured_threads, tokenizer_db, symbolizer)
    )


class ThreadInfo:
    """Provides CPU and stack information that can be inferred from a Thread."""

    _UNKNOWN_VALUE_STR = '0x' + '?' * 8

    def __init__(self, thread: thread_pb2.Thread):
        self._thread = thread

    def _cpu_used_str(self) -> str:
        if not self._thread.HasField('cpu_usage_hundredths'):
            return 'unknown'
        cpu_last_percent = self._thread.cpu_usage_hundredths / 100
        return f'{cpu_last_percent:.2f}%'

    def _stack_size_limit_limit_str(self) -> str:
        if not self.has_stack_size_limit():
            return 'size unknown'

        return f'{self.stack_size_limit()} bytes'

    def _stack_used_str(self) -> str:
        if not self.has_stack_used():
            return 'size unknown'

        used_str = f'{self.stack_used()} bytes'
        if not self.has_stack_size_limit():
            return used_str
        used_str += f', {100*self.stack_used()/self.stack_size_limit():.2f}%'
        return used_str

    def _stack_pointer_est_peak_str(self) -> str:
        if not self.has_stack_pointer_est_peak():
            return 'size unknown'

        high_used_str = f'{self.stack_pointer_est_peak()} bytes'
        if not self.has_stack_size_limit():
            return high_used_str
        high_water_mark_percent = (
            100 * self.stack_pointer_est_peak() / self.stack_size_limit()
        )
        high_used_str += f', {high_water_mark_percent:.2f}%'
        return high_used_str

    def _stack_used_range_str(self) -> str:
        start_str = (
            f'0x{self._thread.stack_start_pointer:08x}'
            if self._thread.HasField('stack_start_pointer')
            else ThreadInfo._UNKNOWN_VALUE_STR
        )
        end_str = (
            f'0x{self._thread.stack_pointer:08x}'
            if self._thread.HasField('stack_pointer')
            else ThreadInfo._UNKNOWN_VALUE_STR
        )

        # TODO(amontanez): Would be nice to represent stack growth direction.
        return f'{start_str} - {end_str} ({self._stack_used_str()})'

    def _stack_limit_range_str(self) -> str:
        start_str = (
            f'0x{self._thread.stack_start_pointer:08x}'
            if self._thread.HasField('stack_start_pointer')
            else ThreadInfo._UNKNOWN_VALUE_STR
        )
        end_str = (
            f'0x{self._thread.stack_end_pointer:08x}'
            if self._thread.HasField('stack_end_pointer')
            else ThreadInfo._UNKNOWN_VALUE_STR
        )

        # TODO(amontanez): Would be nice to represent stack growth direction.
        return f'{start_str} - {end_str} ({self._stack_size_limit_limit_str()})'

    def _stack_pointer_str(self) -> str:
        return (
            f'0x{self._thread.stack_end_pointer:08x}'
            if self._thread.HasField('stack_pointer')
            else ThreadInfo._UNKNOWN_VALUE_STR
        )

    def has_stack_size_limit(self) -> bool:
        """Returns true if there's enough info to calculate stack size."""
        return self._thread.HasField(
            'stack_start_pointer'
        ) and self._thread.HasField('stack_end_pointer')

    def stack_size_limit(self) -> int:
        """Returns the stack size limit in bytes.

        Precondition:
            has_stack_size_limit() must be true.
        """
        assert self.has_stack_size_limit(), 'Missing stack size information'
        return abs(
            self._thread.stack_start_pointer - self._thread.stack_end_pointer
        )

    def has_stack_used(self) -> bool:
        """Returns true if there's enough info to calculate stack usage."""
        return self._thread.HasField(
            'stack_start_pointer'
        ) and self._thread.HasField('stack_pointer')

    def stack_used(self) -> int:
        """Returns the stack usage in bytes.

        Precondition:
            has_stack_used() must be true.
        """
        assert self.has_stack_used(), 'Missing stack usage information'
        return abs(
            self._thread.stack_start_pointer - self._thread.stack_pointer
        )

    def has_stack_pointer_est_peak(self) -> bool:
        """Returns true if there's enough info to calculate estimate
        used stack.
        """
        return self._thread.HasField(
            'stack_start_pointer'
        ) and self._thread.HasField('stack_pointer_est_peak')

    def stack_pointer_est_peak(self) -> int:
        """Returns the max estimated used stack usage in bytes.

        Precondition:
            has_stack_estimated_used_bytes() must be true.
        """
        assert self.has_stack_pointer_est_peak(), 'Missing stack est. peak'
        return abs(
            self._thread.stack_start_pointer
            - self._thread.stack_pointer_est_peak
        )

    def __str__(self) -> str:
        output = [
            f'Est CPU usage: {self._cpu_used_str()}',
            'Stack info',
            f'  Current usage:   {self._stack_used_range_str()}',
            f'  Est peak usage:  {self._stack_pointer_est_peak_str()}',
            f'  Stack limits:    {self._stack_limit_range_str()}',
        ]
        return '\n'.join(output)


class ThreadSnapshotAnalyzer:
    """This class simplifies dumping contents of a snapshot Metadata message."""

    def __init__(
        self,
        threads: thread_pb2.SnapshotThreadInfo,
        tokenizer_db: Optional[pw_tokenizer.Detokenizer] = None,
        symbolizer: Optional[Symbolizer] = None,
    ):
        self._threads = threads.threads
        self._tokenizer_db = (
            tokenizer_db
            if tokenizer_db is not None
            else pw_tokenizer.Detokenizer(None)
        )
        if symbolizer is not None:
            self._symbolizer = symbolizer
        else:
            self._symbolizer = LlvmSymbolizer()

        for thread in self._threads:
            proto_detokenizer.detokenize_fields(self._tokenizer_db, thread)

    def active_thread(self) -> Optional[thread_pb2.Thread]:
        """The thread that requested the snapshot capture."""
        # First check if an interrupt handler was active.
        for thread in self._threads:
            if thread.state == thread_pb2.ThreadState.Enum.INTERRUPT_HANDLER:
                return thread
            if thread.active:  # The deprecated legacy way to report this.
                return thread

        # If not, search for a running thread.
        for thread in self._threads:
            if thread.state == thread_pb2.ThreadState.Enum.RUNNING:
                return thread

        return None

    def __str__(self) -> str:
        """outputs a pw.snapshot.Metadata proto as a multi-line string."""
        output: List[str] = []
        if not self._threads:
            return ''

        output.append('Thread State')
        plural = '' if len(self._threads) == 1 else 's'
        thread_state_overview = f'  {len(self._threads)} thread{plural} running'
        requesting_thread = self.active_thread()
        if not requesting_thread:
            thread_state_overview += '.'
            output.append(thread_state_overview)
        else:
            thread_state_overview += ', '
            underline = ' ' * len(thread_state_overview) + '~' * len(
                requesting_thread.name.decode()
            )
            thread_state_overview += (
                f'{requesting_thread.name.decode()}'
                ' active at the time of capture.'
            )
            output.append(thread_state_overview)
            output.append(underline)

        output.append('')

        # Put the active thread at the front.
        requesting_thread = self.active_thread()
        if requesting_thread is not None:
            self._threads.remove(requesting_thread)
            self._threads.insert(0, requesting_thread)

        for thread in self._threads:
            thread_name = thread.name.decode()
            if not thread_name:
                thread_name = '[unnamed thread]'
            thread_headline = (
                'Thread '
                f'({THREAD_STATE_TO_STRING[thread.state]}): '
                f'{thread_name}'
            )
            if self.active_thread() == thread:
                thread_headline += ' <-- [ACTIVE]'
            output.append(thread_headline)
            output.append(str(ThreadInfo(thread)))
            if thread.raw_backtrace:
                output.append(
                    self._symbolizer.dump_stack_trace(thread.raw_backtrace)
                )
            if thread.raw_stack:
                output.append('Raw Stack')
                output.append(
                    binascii.hexlify(thread.raw_stack, b'\n', 32).decode(
                        'utf-8'
                    )
                )
            # Blank line between threads for nicer formatting.
            output.append('')

        return '\n'.join(output)



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Run an interruptable, cancellable function after debouncing run requests"""

import enum
import logging
import threading
from abc import ABC, abstractmethod

from pw_build.project_builder_context import get_project_builder_context

BUILDER_CONTEXT = get_project_builder_context()

_LOG = logging.getLogger('pw_build.watch')


class DebouncedFunction(ABC):
    """Function to be run by Debouncer"""

    @abstractmethod
    def run(self) -> None:
        """Run the function"""

    @abstractmethod
    def cancel(self) -> bool:
        """Cancel an in-progress run of the function.
        Must be called from different thread than run().
        Returns true if run was successfully cancelled, false otherwise"""

    @abstractmethod
    def on_complete(self, cancelled: bool = False) -> None:
        """Called after run() finishes. If true, cancelled indicates
        cancel() was invoked during the last run()"""

    # Note: The debounce uses threads. Since there is no way to guarantee which
    # thread recieves a KeyboardInterrupt, it is necessary catch this event
    # in all debouncer threads and forward it to the user.
    @abstractmethod
    def on_keyboard_interrupt(self):
        """Called when keyboard interrupt is delivered to a debouncer thread"""


class State(enum.Enum):
    IDLE = 1  # ------- Transistions to: DEBOUNCING
    DEBOUNCING = 2  # - Transistions to: RUNNING
    RUNNING = 3  # ---- Transistions to: INTERRUPTED or COOLDOWN
    INTERRUPTED = 4  # - Transistions to: RERUN
    COOLDOWN = 5  # ---- Transistions to: IDLE
    RERUN = 6  # ------- Transistions to: IDLE (but triggers a press)


class Debouncer:
    """Run an interruptable, cancellable function with debouncing"""

    def __init__(self, function: DebouncedFunction) -> None:
        super().__init__()
        self.function = function

        self.state = State.IDLE

        self.debounce_seconds = 1
        self.debounce_timer = None

        self.cooldown_seconds = 1
        self.cooldown_timer = None

        self.rerun_event_description = ''

        self.lock = threading.Lock()

    def press(self, event_description: str = '') -> None:
        """Try to run the function for the class. If the function is recently
        started, this may push out the deadline for actually starting. If the
        function is already running, will interrupt the function"""
        with self.lock:
            self._press_unlocked(event_description)

    def _press_unlocked(self, event_description: str) -> None:
        _LOG.debug('Press - state = %s', str(self.state))
        if self.state == State.IDLE:
            if event_description:
                _LOG.info('%s', event_description)
            self._start_debounce_timer()
            self._transition(State.DEBOUNCING)

        elif self.state == State.DEBOUNCING:
            self._start_debounce_timer()

        elif self.state == State.RUNNING:
            # When the function is already running but we get an incoming
            # event, go into the INTERRUPTED state to signal that we should
            # re-try running afterwards.
            error_message = ['Event while running: %s', event_description]
            if BUILDER_CONTEXT.using_progress_bars():
                _LOG.warning(*error_message)
            else:
                # Push an empty line to flush ongoing I/O in subprocess.
                print('')

                # Surround the error message with newlines to make it stand out.
                print('')
                _LOG.warning(*error_message)
                print('')

            self.function.cancel()
            self._transition(State.INTERRUPTED)
            self.rerun_event_description = event_description

        elif self.state == State.INTERRUPTED:
            # Function is running but was already interrupted. Do nothing.
            _LOG.debug('Ignoring press - interrupted')

        elif self.state == State.COOLDOWN:
            # Function just finished and we are cooling down; so trigger rerun.
            _LOG.debug('Got event in cooldown; scheduling rerun')
            self._transition(State.RERUN)
            self.rerun_event_description = event_description

    def _transition(self, new_state: State) -> None:
        _LOG.debug('State: %s -> %s', self.state, new_state)
        self.state = new_state

    def _start_debounce_timer(self):
        assert self.lock.locked()
        if self.state == State.DEBOUNCING:
            self.debounce_timer.cancel()
        self.debounce_timer = threading.Timer(
            self.debounce_seconds, self._run_function
        )
        self.debounce_timer.start()

    # Called from debounce_timer thread.
    def _run_function(self):
        try:
            with self.lock:
                assert self.state == State.DEBOUNCING
                self.debounce_timer = None
                self._transition(State.RUNNING)

            # Must run the function without the lock held so further press()
            # calls don't deadlock.
            _LOG.debug('Running debounced function')
            self.function.run()

            _LOG.debug('Finished running debounced function')
            with self.lock:
                if self.state == State.RUNNING:
                    self.function.on_complete(cancelled=False)
                    self._transition(State.COOLDOWN)
                elif self.state == State.INTERRUPTED:
                    self.function.on_complete(cancelled=True)
                    self._transition(State.RERUN)
                self._start_cooldown_timer()
        # Ctrl-C on Unix generates KeyboardInterrupt
        # Ctrl-Z on Windows generates EOFError
        except (KeyboardInterrupt, EOFError):
            self.function.on_keyboard_interrupt()

    def _start_cooldown_timer(self):
        assert self.lock.locked()
        self.cooldown_timer = threading.Timer(
            self.cooldown_seconds, self._exit_cooldown
        )
        self.cooldown_timer.start()

    # Called from cooldown_timer thread.
    def _exit_cooldown(self):
        try:
            with self.lock:
                self.cooldown_timer = None
                rerun = self.state == State.RERUN
                self._transition(State.IDLE)

                # If we were in the RERUN state, then re-trigger the event.
                if rerun:
                    self._press_unlocked(
                        'Rerunning: ' + self.rerun_event_description
                    )

        # Ctrl-C on Unix generates KeyboardInterrupt
        # Ctrl-Z on Windows generates EOFError
        except (KeyboardInterrupt, EOFError):
            self.function.on_keyboard_interrupt()



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and remove optional packages."""

import argparse
import dataclasses
import logging
import os
import pathlib
import shutil
from typing import Dict, List, Optional, Sequence, Tuple

_LOG: logging.Logger = logging.getLogger(__name__)


class Package:
    """Package to be installed.

    Subclass this to implement installation of a specific package.
    """

    def __init__(self, name):
        self._name = name
        self._allow_use_in_downstream = True

    @property
    def name(self):
        return self._name

    @property
    def allow_use_in_downstream(self):
        return self._allow_use_in_downstream

    def install(
        self, path: pathlib.Path
    ) -> None:  # pylint: disable=no-self-use
        """Install the package at path.

        Install the package in path. Cannot assume this directory is empty—it
        may need to be deleted or updated.
        """

    def remove(self, path: pathlib.Path) -> None:  # pylint: disable=no-self-use
        """Remove the package from path.

        Removes the directory containing the package. For most packages this
        should be sufficient to remove the package, and subclasses should not
        need to override this package.
        """
        if os.path.exists(path):
            shutil.rmtree(path)

    def status(  # pylint: disable=no-self-use
        self,
        path: pathlib.Path,  # pylint: disable=unused-argument
    ) -> bool:
        """Returns if package is installed at path and current.

        This method will be skipped if the directory does not exist.
        """
        return False

    def info(  # pylint: disable=no-self-use
        self,
        path: pathlib.Path,  # pylint: disable=unused-argument
    ) -> Sequence[str]:
        """Returns a short string explaining how to enable the package."""
        return []


_PACKAGES: Dict[str, Package] = {}


def register(package_class: type, *args, **kwargs) -> None:
    obj = package_class(*args, **kwargs)
    _PACKAGES[obj.name] = obj


@dataclasses.dataclass
class Packages:
    all: Tuple[str, ...]
    installed: Tuple[str, ...]
    available: Tuple[str, ...]


class UpstreamOnlyPackageError(Exception):
    def __init__(self, pkg_name):
        super().__init__(
            f'Package {pkg_name} is an upstream-only package--it should be '
            'imported as a submodule and not a package'
        )


class PackageManager:
    """Install and remove optional packages."""

    def __init__(self, root: pathlib.Path):
        self._pkg_root = root
        os.makedirs(root, exist_ok=True)

    def install(self, package: str, force: bool = False) -> None:
        """Install the named package.

        Args:
            package: The name of the package to install.
            force: Install the package regardless of whether it's already
                installed or if it's not "allowed" to be installed on this
                project.
        """

        pkg = _PACKAGES[package]
        if not pkg.allow_use_in_downstream:
            if os.environ.get('PW_ROOT') != os.environ.get('PW_PROJECT_ROOT'):
                if force:
                    _LOG.warning(str(UpstreamOnlyPackageError(pkg.name)))
                else:
                    raise UpstreamOnlyPackageError(pkg.name)

        if force:
            self.remove(package)
        pkg.install(self._pkg_root / pkg.name)

    def remove(self, package: str) -> None:
        pkg = _PACKAGES[package]
        pkg.remove(self._pkg_root / pkg.name)

    def status(self, package: str) -> bool:
        pkg = _PACKAGES[package]
        path = self._pkg_root / pkg.name
        return os.path.isdir(path) and pkg.status(path)

    def list(self) -> Packages:
        installed = []
        available = []
        for package in sorted(_PACKAGES.keys()):
            pkg = _PACKAGES[package]
            if pkg.status(self._pkg_root / pkg.name):
                installed.append(pkg.name)
            else:
                available.append(pkg.name)

        return Packages(
            all=tuple(_PACKAGES.keys()),
            installed=tuple(installed),
            available=tuple(available),
        )

    def info(self, package: str) -> Sequence[str]:
        pkg = _PACKAGES[package]
        return pkg.info(self._pkg_root / pkg.name)


class PackageManagerCLI:
    """Command-line interface to PackageManager."""

    def __init__(self):
        self._mgr: PackageManager = None

    def install(self, package: str, force: bool = False) -> int:
        _LOG.info('Installing %s...', package)
        self._mgr.install(package, force)
        _LOG.info('Installing %s...done.', package)
        for line in self._mgr.info(package):
            _LOG.info('%s', line)
        return 0

    def remove(self, package: str) -> int:
        _LOG.info('Removing %s...', package)
        self._mgr.remove(package)
        _LOG.info('Removing %s...done.', package)
        return 0

    def status(self, package: str) -> int:
        if self._mgr.status(package):
            _LOG.info('%s is installed.', package)
            for line in self._mgr.info(package):
                _LOG.info('%s', line)
            return 0

        _LOG.info('%s is not installed.', package)
        return -1

    def list(self) -> int:
        packages = self._mgr.list()

        _LOG.info('Installed packages:')
        for package in packages.installed:
            _LOG.info('  %s', package)
            for line in self._mgr.info(package):
                _LOG.info('    %s', line)
        _LOG.info('')

        _LOG.info('Available packages:')
        for package in packages.available:
            _LOG.info('  %s', package)
        _LOG.info('')

        return 0

    def run(self, command: str, pkg_root: pathlib.Path, **kwargs) -> int:
        self._mgr = PackageManager(pkg_root.resolve())
        return getattr(self, command)(**kwargs)


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser("Manage packages.")
    parser.add_argument(
        '--package-root',
        '-e',
        dest='pkg_root',
        type=pathlib.Path,
        default=pathlib.Path(os.environ['PW_PACKAGE_ROOT']),
    )
    subparsers = parser.add_subparsers(dest='command', required=True)
    install = subparsers.add_parser('install')
    install.add_argument('--force', '-f', action='store_true')
    remove = subparsers.add_parser('remove')
    status = subparsers.add_parser('status')
    for cmd in (install, remove, status):
        cmd.add_argument('package', choices=_PACKAGES.keys())
    _ = subparsers.add_parser('list')
    return parser.parse_args(argv)


def run(**kwargs):
    return PackageManagerCLI().run(**kwargs)



#!/usr/bin/env python

# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Save list of installed packages and versions."""

import argparse
import itertools
import sys
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Union

import pkg_resources


def _installed_packages() -> Iterator[str]:
    """Run pip python_packages and write to out."""
    installed_packages = list(
        pkg.as_requirement()
        for pkg in pkg_resources.working_set  # pylint: disable=not-an-iterable
        # Non-editable packages only
        if isinstance(pkg, pkg_resources.DistInfoDistribution)  # type: ignore
        # This will skip packages with local versions.
        #   For example text after a plus sign: 1.2.3+dev456
        and not pkg.parsed_version.local
        # These are always installed by default in:
        #   pw_env_setup/py/pw_env_setup/virtualenv_setup/install.py
        and pkg.key not in ['pip', 'setuptools', 'wheel']
    )
    for req in sorted(
        installed_packages, key=lambda pkg: pkg.name.lower()  # type: ignore
    ):
        yield str(req)


def ls(output_file: Optional[Path]) -> int:  # pylint: disable=invalid-name
    """Run pip python_packages and write to output_file."""
    actual_requirements = frozenset(
        pkg_resources.Requirement.parse(line) for line in _installed_packages()
    )
    missing_requirements = set()

    # If updating an existing file, load the existing requirements to find lines
    # that are missing in the active environment.
    if output_file:
        existing_lines = output_file.read_text().splitlines()
        expected_requirements = set(
            pkg_resources.Requirement.parse(line) for line in existing_lines
        )
        missing_requirements = expected_requirements - actual_requirements

    new_requirements: List[pkg_resources.Requirement] = list(
        actual_requirements
    )

    for requirement in missing_requirements:
        # Preserve this requirement if it has a marker that doesn't apply to
        # the current environment. For example a line with ;python_version <
        # "3.9" if running on Python 3.9 or higher will be saved.
        if requirement.marker and not requirement.marker.evaluate():
            new_requirements.append(requirement)
            continue

        # If this package is in the active environment then it has a marker
        # in the existing file that should be preserved.
        try:
            found_package = pkg_resources.working_set.find(requirement)
            # If the package version doesn't match, save the new version.
        except pkg_resources.VersionConflict:
            found_package = None
        if found_package:
            # Delete the old line with no marker.
            new_requirements.remove(found_package.as_requirement())
            # Add the existing requirement line that includes the marker.
            new_requirements.append(requirement)

    out = output_file.open('w') if output_file else sys.stdout

    for package in sorted(
        new_requirements, key=lambda pkg: pkg.name.lower()  # type: ignore
    ):
        print(package, file=out)

    if output_file:
        out.close()
    return 0


class UpdateRequiredError(Exception):
    pass


def _stderr(*args, **kwargs):
    return print(*args, file=sys.stderr, **kwargs)


def _load_requirements_lines(*req_files: Path) -> Iterator[str]:
    for req_file in req_files:
        for line in req_file.read_text().splitlines():
            # Ignore comments and blank lines
            if line.startswith('#') or line == '':
                continue
            yield line


def diff(
    expected: Path, ignore_requirements_file: Optional[List[Path]] = None
) -> int:
    """Report on differences between installed and expected versions."""
    actual_lines = set(_installed_packages())
    expected_lines = set(_load_requirements_lines(expected))
    ignored_lines = set()
    if ignore_requirements_file:
        ignored_lines = set(_load_requirements_lines(*ignore_requirements_file))

    if actual_lines == expected_lines:
        _stderr('package versions are identical')
        return 0

    actual_requirements = frozenset(
        pkg_resources.Requirement.parse(line) for line in actual_lines
    )
    expected_requirements = frozenset(
        pkg_resources.Requirement.parse(line) for line in expected_lines
    )
    ignored_requirements = frozenset(
        pkg_resources.Requirement.parse(line) for line in ignored_lines
    )

    removed_requirements = expected_requirements - actual_requirements
    added_requirements = actual_requirements - expected_requirements

    removed_packages: Dict[pkg_resources.Requirement, str] = {}
    updated_packages: Dict[pkg_resources.Requirement, str] = {}
    new_packages: Dict[pkg_resources.Requirement, str] = {}
    reformatted_packages: Dict[pkg_resources.Requirement, str] = {}

    for line in expected_lines:
        requirement = pkg_resources.Requirement.parse(line)

        # Check for lines that need reformatting
        # This will catch lines that use underscores instead of dashes in the
        # name of missing spaces after specifiers.

        # Match this requirement with the original one found in
        # actual_requirements.
        #
        # Note the requirement variable may equal its counterpart in
        # actual_requirements due to the .hashCpm on the Requirement class
        # checking their normalized name. Since we are looking for formatting
        # mismatches here we need to retrieve the requirement instance from the
        # actual_requirements set.
        matching_found_requirement = {
            req: req for req in actual_requirements
        }.get(requirement)

        # If the actual requirement line doesn't match.
        if (
            matching_found_requirement
            and str(matching_found_requirement) != line
        ):
            reformatted_packages[matching_found_requirement] = line

        # If this requirement isn't in the active enviroment and the line
        # doesn't match the repr: flag for reformatting.
        if not matching_found_requirement and str(requirement) != line:
            reformatted_packages[requirement] = line

        # If a requirement specifier is used and it doesn't apply, skip this
        # line. See details for requirement specifiers at:
        # https://pip.pypa.io/en/stable/reference/requirement-specifiers/#requirement-specifiers
        if requirement.marker and not requirement.marker.evaluate():
            continue

        # Try to find this requirement in the current environment.
        try:
            found_package = pkg_resources.working_set.find(requirement)
        # If the package version doesn't match, save the new version.
        except pkg_resources.VersionConflict as err:
            found_package = None
            if err.dist:
                found_package = err.dist

        # If this requirement isn't in the environment, it was removed.
        if not found_package:
            removed_packages[requirement] = line
            continue

        # If found_package is set, the version doesn't match so it was updated.
        if requirement.specs != found_package.as_requirement().specs:
            updated_packages[found_package.as_requirement()] = line

    ignored_distributions = list(
        distribution
        for distribution in pkg_resources.working_set  # pylint: disable=not-an-iterable
        if distribution.as_requirement() in ignored_requirements
    )
    expected_distributions = list(
        distribution
        for distribution in pkg_resources.working_set  # pylint: disable=not-an-iterable
        if distribution.as_requirement() in expected_requirements
    )

    def get_requirements(
        dist_info: pkg_resources.Distribution,
    ) -> Iterator[pkg_resources.Distribution]:
        """Return requirement that are not in expected_distributions."""
        for req in dist_info.requires():
            req_dist_info = pkg_resources.working_set.find(req)
            if not req_dist_info:
                continue
            if req_dist_info in expected_distributions:
                continue
            yield req_dist_info

    def expand_requirements(
        reqs: List[pkg_resources.Distribution],
    ) -> Iterator[List[pkg_resources.Distribution]]:
        """Recursively expand requirements."""
        for dist_info in reqs:
            deps = list(get_requirements(dist_info))
            if deps:
                yield deps
            yield from expand_requirements(deps)

    ignored_transitive_deps = set(
        itertools.chain.from_iterable(
            expand_requirements(ignored_distributions)
        )
    )

    # Check for new packages
    for requirement in added_requirements - removed_requirements:
        if requirement in updated_packages:
            continue
        if requirement in ignored_requirements or ignored_transitive_deps:
            continue

        new_packages[requirement] = str(requirement)

    # Print status messages to stderr

    if reformatted_packages:
        _stderr('Requirements that need reformatting:')
        for requirement, line in reformatted_packages.items():
            _stderr(f'  {line}')
            _stderr('  should be:')
            _stderr(f'  {str(requirement)}')

    if updated_packages:
        _stderr('Updated packages')
        for requirement, line in updated_packages.items():
            _stderr(f'  {str(requirement)} (from {line})')

    if removed_packages:
        _stderr('Removed packages')
        for requirement in removed_packages:
            _stderr(f'  {requirement}')

    if new_packages:
        _stderr('New packages')
        for requirement in new_packages:
            _stderr(f'  {requirement}')

    if updated_packages or new_packages:
        _stderr("Package versions don't match!")
        _stderr(
            f"""
Please do the following:

* purge your environment directory
  * Linux/Mac: 'rm -rf "$_PW_ACTUAL_ENVIRONMENT_ROOT"'
  * Windows: 'rmdir /S %_PW_ACTUAL_ENVIRONMENT_ROOT%'
* bootstrap
  * Linux/Mac: '. ./bootstrap.sh'
  * Windows: 'bootstrap.bat'
* update the constraint file
  * 'pw python-packages list {expected.name}'
"""
        )
        return -1

    return 0


def parse(argv: Union[List[str], None] = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="python -m pw_env_setup.python_packages"
    )
    subparsers = parser.add_subparsers(dest='cmd')

    list_parser = subparsers.add_parser(
        'list', aliases=('ls',), help='List installed package versions.'
    )
    list_parser.add_argument('output_file', type=Path, nargs='?')

    diff_parser = subparsers.add_parser(
        'diff',
        help='Show differences between expected and actual package versions.',
    )
    diff_parser.add_argument('expected', type=Path)
    diff_parser.add_argument(
        '--ignore-requirements-file', type=Path, action='append'
    )

    return parser.parse_args(argv)


def main() -> int:
    args = vars(parse())
    cmd = args.pop('cmd')
    if cmd == 'diff':
        return diff(**args)
    if cmd == 'list':
        return ls(**args)
    return -1


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Configure Python IDE support for Pigweed projects."""

from collections import defaultdict
import os
from pathlib import Path
import platform
from typing import Dict, NamedTuple

_PYTHON_VENV_PATH = (
    Path(os.path.expandvars('$_PW_ACTUAL_ENVIRONMENT_ROOT')) / 'pigweed-venv'
)


class _PythonPathsForPlatform(NamedTuple):
    bin_dir_name: str = 'bin'
    interpreter_name: str = 'python3'


# When given a platform (e.g. the output of platform.system()), this dict gives
# the platform-specific virtualenv path names.
_PYTHON_PATHS_FOR_PLATFORM: Dict[str, _PythonPathsForPlatform] = defaultdict(
    _PythonPathsForPlatform
)
_PYTHON_PATHS_FOR_PLATFORM['Windows'] = _PythonPathsForPlatform(
    bin_dir_name='Scripts', interpreter_name='pythonw.exe'
)


class PythonPaths:
    """Holds the platform-specific Python environment paths.

    The directory layout of Python virtual environments varies among
    platforms. This class holds the data needed to find the right paths
    for a specific platform.
    """

    def __init__(self, system=platform.system()):
        (bin_dir_name, interpreter_name) = _PYTHON_PATHS_FOR_PLATFORM[system]
        self.bin_dir = _PYTHON_VENV_PATH / bin_dir_name
        self.interpreter = self.bin_dir / interpreter_name
